# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gi.repository import GLib
from gi.repository import Gtk

from . import _scrobbler as scrobbler
from . import asprefs
from xl import common, event, xdg, player, settings, providers
from xl.nls import gettext as _
from xlgui.accelerators import Accelerator
from xlgui.widgets import menu
import logging
import time
import pickle
import os

logger = logging.getLogger(__name__)

SCROBBLER = None


def enable(exaile):
    """
    Enables the AudioScrobbler plugin
    """
    global SCROBBLER

    SCROBBLER = ExaileScrobbler(exaile)

    if exaile.loading:
        event.add_callback(__enb, 'gui_loaded')
    else:
        __enb(None, exaile, None)


def __enb(eventname, exaile, nothing):
    GLib.idle_add(_enable, exaile)


def _enable(exaile):
    #    SCROBBLER.exaile_menu = exaile.gui.builder.get_object('tools_menu')
    SCROBBLER.get_options('', '', 'plugin/ascrobbler/menu_check')


def disable(exaile):
    """
    Disables the AudioScrobbler plugin
    """
    global SCROBBLER

    if SCROBBLER:
        SCROBBLER.stop()
        SCROBBLER = None


def get_preferences_pane():
    return asprefs


class ExaileScrobbler:
    def __init__(self, exaile):
        """
        Connects events to the player object, loads settings and cache
        """
        scrobbler.set_user_agent(exaile.get_user_agent_string('audioscrobbler'))
        self.connected = False
        self.connecting = False
        self.use_menu = False
        self.exaile = exaile
        self.cachefile = os.path.join(xdg.get_data_dirs()[0], "audioscrobbler.cache")
        self.get_options('', '', 'plugin/ascrobbler/cache_size')
        self.get_options('', '', 'plugin/ascrobbler/user')
        self.load_cache()
        event.add_ui_callback(self.get_options, 'plugin_ascrobbler_option_set')
        event.add_callback(self._save_cache_cb, 'quit_application')

        # enable accelerator
        def toggle_submit(*x):
            logger.debug('Toggling AudioScrobbler submissions.')
            settings.set_option('plugin/ascrobbler/submit', not self.submit)

        self.accelerator = Accelerator(
            '<Primary>b', _('Toggle AudioScrobbler submit'), toggle_submit
        )
        providers.register('mainwindow-accelerators', self.accelerator)

    def get_options(self, type, sm, option):
        if option == 'plugin/ascrobbler/cache_size':
            self.set_cache_size(
                settings.get_option('plugin/ascrobbler/cache_size', 100), False
            )
            return

        if option in [
            'plugin/ascrobbler/user',
            'plugin/ascrobbler/password',
            'plugin/ascrobbler/submit',
            'plugin/ascrobbler/scrobble_remote',
        ]:
            username = settings.get_option('plugin/ascrobbler/user', '')
            password = settings.get_option('plugin/ascrobbler/password', '')
            server = settings.get_option(
                'plugin/ascrobbler/url', 'http://post.audioscrobbler.com/'
            )
            self.scrobble_remote = settings.get_option(
                'plugin/ascrobbler/scrobble_remote', False
            )
            self.submit = settings.get_option('plugin/ascrobbler/submit', True)

            if (not self.connecting and not self.connected) and self.submit:
                if username and password:
                    self.connecting = True
                    self.initialize(username, password, server)

        if option == 'plugin/ascrobbler/menu_check':
            self.use_menu = settings.get_option('plugin/ascrobbler/menu_check', False)
            if self.use_menu:
                self.setup_menu()
            else:
                self.remove_menu()

    def setup_menu(self):
        self.menu_agr = self.exaile.gui.main.accel_group

        providers.register(
            'menubar-tools-menu',
            menu.simple_separator('plugin-sep', ['track-properties']),
        )

        def factory(menu_, parent, context):
            item = Gtk.CheckMenuItem.new_with_label(_('Enable audioscrobbling'))
            item.set_active(self.submit)
            key, mods = Gtk.accelerator_parse('<Primary>B')
            item.add_accelerator(
                'activate', menu.FAKEACCELGROUP, key, mods, Gtk.AccelFlags.VISIBLE
            )
            item.connect('toggled', self._menu_entry_toggled)
            return item

        item = menu.MenuItem('scrobbler', factory, ['plugin-sep'])
        providers.register('menubar-tools-menu', item)

    def remove_menu(self):
        #        self.menu_entry.disconnect(self.menu_conn)

        for item in providers.get('menubar-tools-menu'):
            if item.name == 'scrobbler':
                providers.unregister('menubar-tools-menu', item)
                break

    def _menu_entry_toggled(self, widget):
        settings.set_option('plugin/ascrobbler/submit', widget.get_active())

    def stop(self):
        """
        Stops submitting
        """
        logger.info("Stopping AudioScrobbler submissions")
        if self.use_menu:
            self.remove_menu()
        if self.connected:
            event.remove_callback(self.on_play, 'playback_track_start', player.PLAYER)
            event.remove_callback(self.on_stop, 'playback_track_end', player.PLAYER)
            self.connected = False
            self.save_cache()
        providers.unregister('mainwindow-accelerators', self.accelerator)

    @common.threaded
    def initialize(self, username, password, server):
        try:
            logger.info("Attempting to connect to AudioScrobbler (%s)", server)
            scrobbler.login(username, password, hashpw=False, post_url=server)
        except Exception:

            try:
                scrobbler.login(username, password, hashpw=True, post_url=server)
            except Exception:
                self.connecting = False
                logger.exception("Error logging in")
                return

        logger.info("Connected to AudioScrobbler")

        event.add_callback(self.on_play, 'playback_track_start', player.PLAYER)
        event.add_callback(self.on_stop, 'playback_track_end', player.PLAYER)
        self.connected = True
        self.connecting = False

    @common.threaded
    def now_playing(self, player, track):
        # wait 5 seconds before now playing to allow for skipping
        time.sleep(5)
        if player.current != track:
            return

        logger.info(
            "Attempting to submit \"Now Playing\" information to AudioScrobbler..."
        )
        try:
            scrobbler.now_playing(
                track.get_tag_raw('artist', join=True),
                track.get_tag_raw('title', join=True),
                track.get_tag_raw('album', join=True),
                int(track.get_tag_raw('__length')),
                track.split_numerical(track.get_tag_raw('tracknumber'))[0] or 0,
            )
        except Exception as e:
            logger.warning("Error submitting \"Now Playing\": %s", e)

    def on_play(self, type, player, track):
        if self.submit:
            track.set_tags(
                __audioscrobbler_playtime=track.get_tag_raw('__playtime'),
                __audioscrobbler_starttime=time.time(),
            )

            if track.is_local() or self.scrobble_remote:
                self.now_playing(player, track)

    def on_stop(self, type, player, track):
        if (
            not track
            or (not track.is_local() and not self.scrobble_remote)
            or track.get_tag_raw('__playtime') is None
        ):
            return
        playtime = (track.get_tag_raw('__playtime') or 0) - (
            track.get_tag_raw('__audioscrobbler_playtime') or 0
        )
        if playtime > 240 or playtime > float(track.get_tag_raw('__length')) / 2.0:
            if self.submit and track.get_tag_raw('__length') > 30:
                self.submit_to_scrobbler(
                    track, track.get_tag_raw('__audioscrobbler_starttime'), playtime
                )

        track.set_tags(__audioscrobbler_starttime=None, __audioscrobbler_playtime=None)

    def set_cache_size(self, size, save=True):
        scrobbler.MAX_CACHE = size
        if save:
            settings.set_option("plugin/ascrobbler/cache_size", size)

    def _save_cache_cb(self, a, b, c):
        self.save_cache()

    def save_cache(self):
        cache = scrobbler.SUBMIT_CACHE
        f = open(self.cachefile, 'wb')
        pickle.dump(cache, f)
        f.close()

    def load_cache(self):
        try:
            f = open(self.cachefile, 'rb')
            cache = pickle.load(f)
            f.close()
            scrobbler.SUBMIT_CACHE = cache
        except Exception:
            pass

    @common.threaded
    def submit_to_scrobbler(self, track, time_started, time_played):
        if scrobbler.SESSION_ID and track and time_started and time_played:
            try:
                scrobbler.submit(
                    track.get_tag_raw('artist', join=True),
                    track.get_tag_raw('title', join=True),
                    int(time_started),
                    'P',
                    '',
                    int(track.get_tag_raw('__length')),
                    track.get_tag_raw('album', join=True),
                    track.split_numerical(track.get_tag_raw('tracknumber'))[0] or 0,
                    autoflush=True,
                )
            except Exception:
                logger.exception("AS: Failed to submit track")



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from collections import namedtuple
import copy
import threading
from typing import Any, ClassVar, Mapping, Sequence

import logging

logger = logging.getLogger(__name__)

from xl import version
import mutagen

version.register('Mutagen', mutagen.version_string)


INFO_TAGS = ['__bitrate', '__length']

_compute_lock = threading.Lock()

# Generic description of cover images
# - type is a number corresponding to the cover type of ID3 APIC tags,
#   desc is a string describing the image, mime is a type,
#   data is the img data
# -> if type is None, then the type is not changeable
CoverImage = namedtuple('CoverImage', 'type desc mime data')


class NotWritable(Exception):
    pass


class NotReadable(Exception):
    pass


class BaseFormat:
    """
    Base class for handling loading of metadata from files.

    subclasses using mutagen should set MutagenType and overload
    the _get_tag, _set_tag, and _del_tag methods as needed.

    subclasses not using mutagen should leave MutagenType as None
    """

    MutagenType = None

    # This should contain ALL keys supported by this filetype, unless 'others'
    # is set to True. If others is True, then anything not in tag_mapping will
    # be written using the Exaile tag name
    # -> dict k: exaile tag name, v: native tag name
    tag_mapping: ClassVar[Mapping[str, Any]] = {}
    others: ClassVar[bool] = True
    writable: ClassVar[bool] = False
    case_sensitive: ClassVar[bool] = True

    # Whether _compute_mappings has run
    _computed: ClassVar[bool]
    # Reverse of tag_mapping, populated in _compute_mappings
    _reverse_mapping: ClassVar[Mapping[Any, str]]

    @classmethod
    def _compute_mappings(cls):
        with _compute_lock:
            # this only needs to be run once per class
            if hasattr(cls, '_computed'):
                return

            if cls.case_sensitive:
                cls._reverse_mapping = {v: k for k, v in cls.tag_mapping.items()}
            else:
                cls._reverse_mapping = {
                    v.lower(): k for k, v in cls.tag_mapping.items()
                }

            from .tags import disk_tags

            cls.ignore_tags = set(disk_tags)

            # This comes last to indicate success
            cls._computed = True

    def __init__(self, loc):
        """
        Raises :class:`NotReadable` if the file cannot be
        opened for some reason.

        :param loc: absolute path to the file to read
            (note - this may change to accept gio uris in the future)
        """
        self.loc = loc
        self.open = False
        self.mutagen = None
        try:
            self._computed
        except AttributeError:
            self.__class__._compute_mappings()

        self.load()

    def load(self):
        """
        Loads the tags from the file.
        """
        if self.MutagenType:
            try:
                self.mutagen = self.MutagenType(self.loc)
            except Exception:
                raise NotReadable

    def save(self):
        """
        Saves any changes to the tags.
        """
        if self.writable and self.mutagen:
            self.mutagen.save()

    def _del_tag(self, raw, tag):
        """
        :param tag: The native tag name
        """
        if tag in raw:
            del raw[tag]

    def _get_raw(self):
        if self.MutagenType or self.mutagen:
            return self.mutagen
        else:
            return {}

    def _get_tag(self, raw, tag):
        """
        :param tag: The native tag name
        """
        try:
            return raw[tag]
        except KeyError:
            return None

    def _set_tag(self, raw, tag, value):
        """
        :param tag: The native tag name
        :param value: If None, delete the tag
        """
        raw[tag] = value

    def get_keys_disk(self) -> Sequence[str]:
        """
        Returns keys of all tags that can be read from disk
        """
        return [self._reverse_mapping.get(k, k) for k in self._get_raw().keys()]

    def read_all(self):
        """
        Reads all non-blacklisted tags from the file.

        Blacklisted tags include lyrics, covers, and any field starting
        with __. If you need to read these, call read_tags directly.
        """
        tags = INFO_TAGS[:]
        for t in self.get_keys_disk():
            if t in self.ignore_tags:
                continue
            # __ is used to denote exaile's internal tags, so we skip
            # loading them to avoid conflicts. usually this shouldn't be
            # an issue.
            if t.startswith("__"):
                continue
            tags.append(t)
        alltags = self.read_tags(tags)
        return alltags

    def read_tags(self, tags):
        """
        get the values for the specified tags.

        returns a dict of the found values. if no value was found for a
        requested tag it will not exist in the returned dict.

        :param tags: a list of exaile tag names to read
        :returns: a dictionary of tag/value pairs.
        """
        raw = self._get_raw()
        td = {}
        for tag in tags:
            t = None
            if tag in INFO_TAGS:
                try:
                    t = self.get_info(tag)
                except KeyError:
                    pass
            if t is None and tag in self.tag_mapping:
                try:
                    t = self._get_tag(raw, self.tag_mapping[tag])
                    if isinstance(t, (str, bytes)):
                        t = [t]
                    elif isinstance(t, list):
                        pass
                    elif t is not None:
                        t = [str(u) for u in t]
                except (KeyError, TypeError):
                    logger.debug("Unexpected error reading `%s`", tag, exc_info=True)
            if t is None and self.others and tag not in self.tag_mapping:
                try:
                    t = self._get_tag(raw, tag)
                    if isinstance(t, (str, bytes)):
                        t = [t]
                    elif t is not None:
                        t = [str(u) for u in t]
                except (KeyError, TypeError):
                    logger.debug("Unexpected error reading `%s`", tag, exc_info=True)

            if t:
                td[tag] = t
        return td

    def write_tags(self, tagdict):
        """
        Write a set of tags to the file. Raises a NotWritable exception
        if the format does not support writing tags.

        When calling this function, we assume the following:

        * tagdict has all keys that you wish to write, keys are exaile tag
          names or custom tag names and values are the tags to write (lists
          of unicode strings)
        * if a value is None, then that tag will be deleted from the file
        * Will not modify/delete tags that are NOT in tagdict
        * Will not write tags that start with '__'

        :param tagdict: A dictionary of tag/value pairs to write.
        """
        if not self.MutagenType or not self.writable:
            raise NotWritable
        else:
            tagdict = copy.deepcopy(tagdict)
            raw = self._get_raw()
            # Add tags if it doesn't have them. Mutagen will throw an exception
            # if the file already contains tags.
            if getattr(raw, 'tags', None) is None:
                try:
                    raw.add_tags()
                except Exception:
                    # XXX: Not sure needed since we're already checking for
                    # existence of tags.
                    pass

            # tags starting with __ are internal and should not be written
            # -> this covers INFO_TAGS, which also shouldn't be written
            for tag in list(tagdict.keys()):
                if tag.startswith("__"):
                    del tagdict[tag]

            # Only modify the tags we were told to modify
            # -> if the value is None, delete the tag
            for tag, value in tagdict.items():
                rtag = self.tag_mapping.get(tag)
                if rtag:
                    if value is not None:
                        self._set_tag(raw, rtag, value)
                    else:
                        self._del_tag(raw, rtag)
                elif self.others:
                    if value is not None:
                        self._set_tag(raw, tag, value)
                    else:
                        self._del_tag(raw, tag)

            self.save()

    def get_info(self, info):
        # TODO: add sample rate? filesize?
        if info == "__length":
            return self.get_length()
        elif info == "__bitrate":
            return self.get_bitrate()
        else:
            raise KeyError

    def get_length(self):
        try:
            return self.mutagen.info.length
        except AttributeError:
            try:
                return self._get_raw()['__length']
            except (KeyError, TypeError):
                return None

    def get_bitrate(self):
        try:
            return self.mutagen.info.bitrate
        except AttributeError:
            try:
                return self._get_raw()['__bitrate']
            except (KeyError, TypeError):
                return None


class CaseInsensitveBaseFormat(BaseFormat):
    case_sensitive = False

    def get_keys_disk(self):
        """
        Returns keys of all tags that can be read from disk
        """
        return [self._reverse_mapping.get(k.lower(), k) for k in self._get_raw().keys()]


# vim: et sts=4 sw=4



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _('Alarm Clock')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "acprefs_pane.ui")


class HourPreference(widgets.SpinPreference):
    default = 12
    name = 'plugin/alarmclock/hour'


class MinutesPreference(widgets.SpinPreference):
    default = 30
    name = 'plugin/alarmclock/minuts'


class MondayPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/alarmclock/monday'


class TuesdayPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/alarmclock/tuesday'


class WednesdayPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/alarmclock/wednesday'


class ThursdayPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/alarmclock/thursday'


class FridayPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/alarmclock/friday'


class SaturdayPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/alarmclock/saturday'


class SundayPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/alarmclock/sunday'


class FadingPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/alarmclock/alarm_use_fading'


class MinVolumePreference(widgets.SpinPreference):
    default = 0
    name = 'plugin/alarmclock/alarm_min_volume'


class MaxVolumePreference(widgets.SpinPreference):
    default = 100
    name = 'plugin/alarmclock/alarm_max_volume'


class IncrementPreference(widgets.SpinPreference):
    default = 1
    name = 'plugin/alarmclock/alarm_increment'


class TimerStepPreference(widgets.SpinPreference):
    default = 1
    name = 'plugin/alarmclock/alarm_timer_per_inc'



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gst

from xl.nls import gettext as _
import logging

logger = logging.getLogger(__name__)

"""
    explanation of format dicts:
    default:    the default quality to use, must be a member of raw_steps.
    raw_steps:  a value defining the quality of encoding that will be passed
                to the encoder.
    kbs_steps:  a value defining the quality of encoding that will be displayed
                to the user. must be a one-to-one mapping with raw_steps.
    command:    the gstreamer pipeline to execute. should contain exactly one
                python string format operator, like %s or %f, which will be
                replaced with the value from raw_steps.
    extension:  the default filename extension for this format
    plugins:    the gstreamer plugins needed for this transcode pipeline
    desc:       a description of the encoder to display to the user
"""

FORMATS = {
    # fmt: off
    "Ogg Vorbis" : {
        "default"   : 0.5,
        "raw_steps" : [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
        "kbs_steps" : [64, 80, 96, 112, 128, 160, 192, 224, 256, 320],
        "command"   : "vorbisenc quality=%1.1f ! oggmux",
        "extension" : "ogg",
        "plugins"   : ["vorbisenc", "oggmux"],
        "desc"      : _("Vorbis is an open source, lossy audio codec with "
                        "high quality output at a lower file size than MP3.")
    },
    "FLAC" : {
        "default"   : 5,
        "raw_steps" : [0, 1, 2, 3, 4, 5, 6, 7, 8],
        "kbs_steps" : [0, 1, 2, 3, 4, 5, 6, 7, 8],
        "command"   : "flacenc quality=%i",
        "extension" : "flac",
        "plugins"   : ["flacenc"],
        "desc"      : _("Free Lossless Audio Codec (FLAC) is an open "
                        "source codec that compresses but does not degrade audio "
                        "quality.")
    },
    "AAC"       : {
        "default"   : 160000,
        "raw_steps" : [32000, 48000, 64000, 96000, 128000, 160000,
                       192000, 224000, 256000, 320000],
        "kbs_steps" : [32, 48, 64, 96, 128, 160, 192, 224, 256, 320],
        "command"   : "faac bitrate=%i ! ffmux_mp4",
        "extension" : "m4a",
        "plugins"   : ["faac", "ffmux_mp4"],
        "desc"      : _("Apple's proprietary lossy audio format that "
                        "achieves better sound quality than MP3 at "
                        "lower bitrates.")
    },
    "MP3 (VBR)" : {
        "default"   : 160,
        "raw_steps" : [32, 48, 64, 96, 128, 160, 192, 224, 256, 320],
        "kbs_steps" : [32, 48, 64, 96, 128, 160, 192, 224, 256, 320],
        "command"   : "lame vbr=4 vbr-mean-bitrate=%i",
        "extension" : "mp3",
        "plugins"   : ["lamemp3enc"],
        "desc"      : _("A proprietary and older, but also popular, lossy "
                        "audio format. VBR gives higher quality than CBR, but may "
                        "be incompatible with some players.")
    },
    "MP3 (CBR)" : {
        "default"   : 160,
        "raw_steps" : [32, 48, 64, 96, 128, 160, 192, 224, 256, 320],
        "kbs_steps" : [32, 48, 64, 96, 128, 160, 192, 224, 256, 320],
        "command"   : "lame bitrate=%i",
        "extension" : "mp3",
        "plugins"   : ["lamemp3enc"],
        "desc"      : _("A proprietary and older, but also popular, "
                        "lossy audio format. CBR gives less quality than VBR, "
                        "but is compatible with any player.")
    },
    "WavPack" : {
        "default"   : 2,
        "raw_steps" : [1, 2, 3, 4],
        "kbs_steps" : [1, 2, 3, 4],
        "command"   : "wavpackenc mode=%i",
        "extension" : "wv",
        "plugins"   : ["wavpackenc"],
        "desc"      : _("A very fast Free lossless audio format with "
                        "good compression."),
    },
    # fmt: on
}

# NOTE: the transcoder is NOT designed to transfer tags. You will need to
# manually write the tags after transcoding has completed.


def get_formats():
    ret = {}
    for name, val in FORMATS.items():
        try:
            for plug in val['plugins']:
                x = Gst.ElementFactory.find(plug)
                if not x:
                    raise
            ret[name] = val
        except Exception:
            pass
    return ret


def add_format(name, fmt):
    global FORMATS
    FORMATS[name] = fmt


class TranscodeError(Exception):
    pass


class Transcoder:
    def __init__(self, destformat, quality, error_callback, end_callback):
        self.src = None
        self.sink = None
        self.set_format(destformat)
        self.set_quality(quality)
        self.input = None
        self.output = None
        self.encoder = None
        self.pipe = None
        self.bus = None
        self.running = False
        self.__last_time = 0.0
        self.error_cb = error_callback
        self.end_cb = end_callback

    def set_format(self, name):
        if name in FORMATS:
            self.dest_format = name

    def set_quality(self, value):
        if value in FORMATS[self.dest_format]['raw_steps']:
            self.quality = value

    def _construct_encoder(self):
        fmt = FORMATS[self.dest_format]
        quality = self.quality
        self.encoder = fmt["command"] % quality

    def set_input(self, uri):
        self.input = """filesrc location="%s" """ % uri

    def set_raw_input(self, raw):
        self.input = raw

    def set_output(self, uri):
        self.output = """filesink location="%s" """ % uri

    def set_output_raw(self, raw):
        self.output = raw

    def start_transcode(self):
        self._construct_encoder()
        elements = [
            self.input,
            "decodebin name=\"decoder\"",
            "audioconvert",
            self.encoder,
            self.output,
        ]
        pipestr = " ! ".join(elements)
        logger.info("Starting GStreamer decoder with pipestring: %s", pipestr)
        pipe = Gst.parse_launch(pipestr)
        self.pipe = pipe
        self.bus = pipe.get_bus()
        self.bus.add_signal_watch()
        self.bus.connect('message::error', self.on_error)
        self.bus.connect('message::eos', self.on_eof)

        pipe.set_state(Gst.State.PLAYING)
        self.running = True
        return pipe

    def stop(self):
        self.pipe.set_state(Gst.State.NULL)
        self.running = False
        self.__last_time = 0.0
        self.end_cb()

    def on_error(self, bus, message):
        self.pipe.set_state(Gst.State.NULL)
        self.running = False
        gerror, message_string = message.parse_error()
        self.error_cb(gerror, message_string)
        logger.error(message_string)
        raise gerror

    def on_eof(self, bus, message):
        self.stop()

    def get_time(self):
        if not self.running:
            return 0.0
        try:
            tim = self.pipe.query_position(Gst.Format.TIME)[0]
            tim = tim / Gst.SECOND
            self.__last_time = tim
            return tim
        except Exception:
            return self.__last_time

    def is_running(self):
        return self.running



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from xl.metadata._base import BaseFormat

import logging
import os


logger = logging.getLogger(__name__)


try:
    import ctypes

    modplug = ctypes.cdll.LoadLibrary("libmodplug.so.1")
    modplug.ModPlug_Load.restype = ctypes.c_void_p
    modplug.ModPlug_Load.argtypes = (ctypes.c_void_p, ctypes.c_int)
    modplug.ModPlug_GetName.restype = ctypes.c_char_p
    modplug.ModPlug_GetName.argtypes = (ctypes.c_void_p,)
    modplug.ModPlug_GetLength.restype = ctypes.c_int
    modplug.ModPlug_GetLength.argtypes = (ctypes.c_void_p,)
except (ImportError, OSError):
    logger.debug('No support for Mod metadata because libmodplug could not be found.')
    modplug = None


class ModFormat(BaseFormat):
    writable = False

    def load(self):
        if modplug:
            data = open(self.loc, "rb").read()
            f = modplug.ModPlug_Load(data, len(data))
            if f:
                name = modplug.ModPlug_GetName(f) or os.path.split(self.loc)[-1]
                length = modplug.ModPlug_GetLength(f) / 1000.0 or -1
                self.mutagen = {'title': name, '__length': length}
        else:
            self.mutagen = {}

    def get_length(self):
        try:
            return self.mutagen['__length']
        except KeyError:
            return -1

    def get_bitrate(self):
        return -1


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
    General functions and classes shared in the codebase
"""

from collections import deque
import collections.abc
from functools import wraps, partial
import inspect
import logging
import os
import os.path
import pickle
import shelve
import subprocess
import sys
import threading
import urllib.parse
import urllib.request
import weakref

import bsddb3 as bsddb
from gi.repository import Gio, GLib, GObject

from xl import shelve_compat


logger = logging.getLogger(__name__)


# TODO: get rid of this. only plugins/cd/ uses it.
VALID_TAGS = (
    # Ogg Vorbis spec tags
    "title version album tracknumber artist genre performer copyright "
    "license organization description location contact isrc date "
    # Other tags we like
    "arranger author composer conductor lyricist discnumber labelid part "
    "website language encodedby bpm albumartist originaldate originalalbum "
    "originalartist recordingdate"
).split()

PICKLE_PROTOCOL = 2

# Default tags for track sorting. Unless you have good reason to do
# otherwise, use this.
# TODO: make this a setting?
BASE_SORT_TAGS = ('albumartist', 'date', 'album', 'discnumber', 'tracknumber', 'title')


def clamp(value, minimum, maximum):
    """
    Clamps a value to the given boundaries

    :param value: the value to clamp
    :param minimum: the minimum value to return
    :param maximum: the maximum value to return
    """
    return max(minimum, min(value, maximum))


def enum(**enums):
    """
    Creates an enum type

    :see: https://stackoverflow.com/a/1695250
    """
    return type('Enum', (), enums)


def sanitize_url(url):
    """
    Removes the password part from an url

    :param url: the URL to sanitize
    :type url: string
    :returns: the sanitized url
    """
    try:
        components = list(urllib.parse.urlparse(url))
        auth, host = components[1].split('@')
        username, password = auth.split(':')
    except (AttributeError, ValueError):
        pass
    else:
        # Replace password with fixed amount of "*"
        auth = ':'.join((username, 5 * '*'))
        components[1] = '@'.join((auth, host))
        url = urllib.parse.urlunparse(components)

    return url


def get_url_contents(url, user_agent):
    """
    Retrieves data from a URL and sticks a user-agent on it. You can use
    exaile.get_user_agent_string(pluginname) to get this.

    Added in Exaile 3.4

    :returns: Contents of page located at URL
    :raises: urllib.error.URLError
    """

    headers = {'User-Agent': user_agent}
    req = urllib.request.Request(url, None, headers)
    fp = urllib.request.urlopen(req)
    data = fp.read()
    fp.close()

    return data


def threaded(func):
    """
    A decorator that will make any function run in a new thread

    :param func: the function to run threaded
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        t = threading.Thread(target=func, args=args, kwargs=kwargs)
        t.daemon = True
        t.start()

    return wrapper


def synchronized(func):
    """
    A decorator to make a function synchronized - which means only one
    thread is allowed to access it at a time.

    This only works on class functions, and creates a variable in
    the instance called _sync_lock.

    If this function is used on multiple functions in an object, they
    will be locked with respect to each other. The lock is re-entrant.
    """

    @wraps(func)
    def wrapper(self, *__args, **__kw):
        try:
            rlock = self._sync_lock
        except AttributeError:
            from threading import RLock

            rlock = self.__dict__.setdefault('_sync_lock', RLock())
        rlock.acquire()
        try:
            return func(self, *__args, **__kw)
        finally:
            rlock.release()

    return wrapper


def _idle_callback(func, callback, *args, **kwargs):
    value = func(*args, **kwargs)
    if callback and callable(callback):
        callback(value)


def idle_add(callback=None):
    """
    A decorator that will wrap the function in a GLib.idle_add call

    NOTE: Although this decorator will probably work in more cases than
    the gtkrun decorator does, you CANNOT expect to get a return value
    from the function that calls a function with this decorator.  Instead,
    you must use the callback parameter.  If the wrapped function returns
    a value, it will be passed in as a parameter to the callback function.

    @param callback: optional callback that will be called when the
        wrapped function is done running
    """

    def wrap(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            GLib.idle_add(_idle_callback, f, callback, *args, **kwargs)

        return wrapped

    return wrap


def _glib_wait_inner(timeout, glib_timeout_func):
    # Have to hold the value in a mutable structure because python's scoping
    # rules prevent us assigning to an outer scope directly.
    #
    # Additionally, we hold source ids per-instance, otherwise this would
    # restrict calls across all instances of an object with the glib_wait*
    # decorators, which would have surprising results
    id_by_obj = weakref.WeakKeyDictionary()

    def waiter(function):
        # ensure this is only used on instance methods
        callargs = inspect.getfullargspec(function)
        if len(callargs.args) == 0 or callargs.args[0] != 'self':
            raise RuntimeError("Must only use glib_wait* on instance methods!")

        def thunk(*args, **kwargs):
            id_by_obj[args[0]] = None
            # if a function returns True, it wants to be called again; in that
            # case, treat it as an additional call, otherwise you can potentially
            # get lots of callbacks piling up
            if function(*args, **kwargs):
                delayer(*args, **kwargs)

        def delayer(*args, **kwargs):
            self = args[0]
            srcid = id_by_obj.get(self)
            if srcid:
                GLib.source_remove(srcid)
            id_by_obj[self] = glib_timeout_func(timeout, thunk, *args, **kwargs)

        return delayer

    return waiter


def glib_wait(timeout):
    """
    Decorator to make a function run only after 'timeout'
    milliseconds have elapsed since the most recent call to the
    function.

    For example, if a function was given a timeout of 1000 and
    called once, then again half a second later, it would run
    only once, 1.5 seconds after the first call to it.

    If arguments are given to the function, only the last call's set
    of arguments will be used.

    If the function returns a value that evaluates to True, it
    will be called again under the same timeout rules.

    .. warning:: Can only be used with instance methods
    """
    return _glib_wait_inner(timeout, GLib.timeout_add)


def glib_wait_seconds(timeout):
    """
    Same as glib_wait, but uses GLib.timeout_add_seconds instead
    of GLib.timeout_add and takes its timeout in seconds. See the
    glib documention for why you might want to use one over the
    other.
    """
    return _glib_wait_inner(timeout, GLib.timeout_add_seconds)


def profileit(func):
    """
    Decorator to profile a function
    """
    import cProfile
    import pstats

    @wraps(func)
    def wrapper(*args, **kwargs):
        prof = cProfile.Profile()
        res = prof.runcall(func, *args, **kwargs)
        stats = pstats.Stats(prof)
        stats.strip_dirs()
        stats.sort_stats('time', 'calls')
        print(">>>---- Begin profiling print")
        stats.print_stats()
        print(">>>---- End profiling print")
        prof = stats = None
        return res

    return wrapper


class classproperty:
    """
    Decorator allowing for class property access
    """

    def __init__(self, function):
        self.function = function

    def __get__(self, obj, type):
        return self.function(type)


class VersionError(Exception):
    """
    Represents version discrepancies
    """

    #: the error message
    message = None

    def __init__(self, message):
        Exception.__init__(self)
        self.message = message

    def __str__(self):
        return repr(self.message)


def open_file(path):
    """
    Opens a file or folder using the system configured program
    """
    platform = sys.platform
    if platform == 'win32':
        # pylint will error here on non-windows platforms unless we do this
        # pylint: disable-msg=E1101
        os.startfile(path)
        # pylint: enable-msg=E1101
    elif platform == 'darwin':
        subprocess.Popen(["open", path])
    else:
        subprocess.Popen(["xdg-open", path])


def open_file_directory(path_or_uri):
    """
    Opens the parent directory of a file, selecting the file if possible.
    """
    f = Gio.File.new_for_commandline_arg(path_or_uri)
    platform = sys.platform
    if platform == 'win32':
        # We could run `explorer /select,filename`, but that doesn't support
        # reusing an existing Explorer window when selecting a file in a
        # directory that is already open.

        import ctypes

        CoInitialize = ctypes.windll.ole32.CoInitialize
        CoInitialize.argtypes = [ctypes.c_void_p]
        CoInitialize.restype = ctypes.HRESULT
        CoUninitialize = ctypes.windll.ole32.CoUninitialize
        CoUninitialize.argtypes = []
        CoUninitialize.restype = None
        ILCreateFromPath = ctypes.windll.shell32.ILCreateFromPathW
        ILCreateFromPath.argtypes = [ctypes.c_wchar_p]
        ILCreateFromPath.restype = ctypes.c_void_p
        ILFree = ctypes.windll.shell32.ILFree
        ILFree.argtypes = [ctypes.c_void_p]
        ILFree.restype = None
        SHOpenFolderAndSelectItems = ctypes.windll.shell32.SHOpenFolderAndSelectItems
        SHOpenFolderAndSelectItems.argtypes = [
            ctypes.c_void_p,
            ctypes.c_uint,
            ctypes.c_void_p,
            ctypes.c_ulong,
        ]
        SHOpenFolderAndSelectItems.restype = ctypes.HRESULT

        CoInitialize(None)
        pidl = ILCreateFromPath(f.get_path())
        SHOpenFolderAndSelectItems(pidl, 0, None, 0)
        ILFree(pidl)
        CoUninitialize()
    elif platform == 'darwin':
        subprocess.Popen(["open", f.get_parent().get_parse_name()])
    else:
        subprocess.Popen(["xdg-open", f.get_parent().get_parse_name()])


def open_shelf(path):
    """
    Opens a python shelf file, used to store various types of metadata
    """
    shelve_compat.ensure_shelve_compat()

    # As of Exaile 4, new DBs will only be created as Berkeley DB Hash databases
    # using either bsddb3 (external) or bsddb (stdlib but sometimes removed).
    # Existing DBs created with other backends will be migrated to Berkeley DB.
    # We do this because BDB is generally considered more performant,
    # and because gdbm currently doesn't work at all in MSYS2.

    # Some DBM modules don't use the path we give them, but rather they have
    # multiple filenames. If the specified path doesn't exist, double check
    # to see if whichdb returns a result before trying to open it with bsddb
    force_migrate = False
    if not os.path.exists(path):
        from dbm import whichdb

        if whichdb(path) is not None:
            force_migrate = True

    if not force_migrate:
        try:
            db = bsddb.hashopen(path, 'c')
            return shelve.BsdDbShelf(db, protocol=PICKLE_PROTOCOL)
        except bsddb.db.DBInvalidArgError:
            logger.warning("%s was created with an old backend, migrating it", path)
        except Exception:
            raise

    # special case: zero-length file
    if not force_migrate and os.path.getsize(path) == 0:
        os.unlink(path)
    else:
        from xl.migrations.database.to_bsddb import migrate

        migrate(path)

    db = bsddb.hashopen(path, 'c')
    return shelve.BsdDbShelf(db, protocol=PICKLE_PROTOCOL)


class LimitedCache(collections.abc.MutableMapping):
    """
    Simple cache that acts much like a dict, but has a maximum # of items
    """

    def __init__(self, limit):
        self.limit = limit
        self.order = deque()
        self.cache = dict()

    def __iter__(self):
        return self.cache.__iter__()

    def __len__(self):
        return self.cache.__len__()

    def __contains__(self, item):
        return self.cache.__contains__(item)

    def __delitem__(self, item):
        del self.cache[item]
        self.order.remove(item)

    def __getitem__(self, item):
        val = self.cache[item]
        self.order.remove(item)
        self.order.append(item)
        return val

    def __setitem__(self, item, value):
        self.cache[item] = value
        if item in self.order:
            self.order.remove(item)
        self.order.append(item)
        while len(self) > self.limit:
            del self.cache[self.order.popleft()]

    def __repr__(self):
        '''prevent repr(self) from changing cache order'''
        return repr(self.cache)

    def __str__(self):
        '''prevent str(self) from changing cache order'''
        return str(self.cache)

    def keys(self):
        return self.cache.keys()


class cached:
    """
    Decorator to make a function's results cached
    does not cache if there is an exception.

    .. note:: This probably breaks on functions that modify their arguments
    """

    def __init__(self, limit):
        self.limit = limit

    @staticmethod
    def _freeze(d):
        return frozenset(d.items())

    def __call__(self, f):
        try:
            f._cache
        except AttributeError:
            f._cache = LimitedCache(self.limit)

        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f._cache[(args, self._freeze(kwargs))]
            except KeyError:
                pass
            ret = f(*args, **kwargs)
            try:
                f._cache[(args, self._freeze(kwargs))] = ret
            except TypeError:  # args can't be hashed
                pass
            return ret

        return wrapper

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return partial(self.__call__, obj)


def walk(root):
    """
    Walk through a Gio directory, yielding each file

    Files are enumerated in the following order: first the
    directory, then the files in that directory. Once one
    directory's files have all been listed, it moves on to
    the next directory. Order of files within a directory
    and order of directory traversal is not specified.

    :param root: a :class:`Gio.File` representing the
        directory to walk through
    :returns: a generator object
    :rtype: :class:`Gio.File`
    """
    queue = deque()
    queue.append(root)

    while len(queue) > 0:
        dir = queue.pop()
        yield dir
        try:
            for fileinfo in dir.enumerate_children(
                "standard::type,"
                "standard::is-symlink,standard::name,"
                "standard::symlink-target,time::modified",
                Gio.FileQueryInfoFlags.NONE,
                None,
            ):
                fil = dir.get_child(fileinfo.get_name())
                # FIXME: recursive symlinks could cause an infinite loop
                if fileinfo.get_is_symlink():
                    target = fileinfo.get_symlink_target()
                    if "://" not in target and not os.path.isabs(target):
                        fil2 = dir.get_child(target)
                    else:
                        fil2 = Gio.File.new_for_uri(target)
                    # already in the collection, we'll get it anyway
                    if fil2.has_prefix(root):
                        continue
                type = fileinfo.get_file_type()
                if type == Gio.FileType.DIRECTORY:
                    queue.append(fil)
                elif type == Gio.FileType.REGULAR:
                    yield fil
        except GLib.Error:  # why doesnt gio offer more-specific errors?
            logger.exception("Unhandled exception while walking on %s.", dir)


def walk_directories(root):
    """
    Walk through a Gio directory, yielding each subdirectory

    :param root: a :class:`Gio.File` representing the
        directory to walk through
    :returns: a generator object
    :rtype: :class:`Gio.File`
    """
    yield root
    directory = None
    subdirectory = None

    try:
        for fileinfo in root.enumerate_children(
            'standard::name,standard::type', Gio.FileQueryInfoFlags.NONE, None
        ):
            if fileinfo.get_file_type() == Gio.FileType.DIRECTORY:
                directory = root.get_child(fileinfo.get_name())

                for subdirectory in walk_directories(directory):
                    yield subdirectory
    except GLib.Error:
        logger.exception(
            "Unhandled exception while walking dirs on %s, %s, %s",
            root,
            directory,
            subdirectory,
        )


class TimeSpan:
    """
    Calculates the number of days, hours, minutes,
    and seconds in a time span
    """

    #: number of days
    days = 0
    #: number of hours
    hours = 0
    #: number of minutes
    minutes = 0
    #: number of seconds
    seconds = 0

    def __init__(self, span):
        """
        :param span: Time span in seconds
        :type span: float
        """
        try:
            span = float(span)
        except (ValueError, TypeError):
            span = 0

        span, self.seconds = divmod(span, 60)
        span, self.minutes = divmod(span, 60)
        self.days, self.hours = divmod(span, 24)

    def __repr__(self):
        span = self.days * 24 + self.hours
        span = span * 60 + self.minutes
        span = span * 60 + self.seconds
        return '%s(%s)' % (self.__class__.__name__, span)

    def __str__(self):
        return '%dd, %dh, %dm, %ds' % (
            self.days,
            self.hours,
            self.minutes,
            self.seconds,
        )


class MetadataList:
    """
    Like a list, but also associates an object of metadata
    with each entry.

    ``(get|set|del)_meta_key`` are the metadata interface - they
    allow the metadata to act much like a dictionary, with a few
    optimizations.

    List aspects that are not supported:
        * sort
        * comparisons other than equality
        * multiply
    """

    __slots__ = ['__list', 'metadata']

    def __init__(self, iterable=[], metadata=[]):
        self.__list = list(iterable)
        meta = list(metadata)
        if meta and len(meta) != len(self.__list):
            raise ValueError("Length of metadata must match length of items.")
        if not meta:
            meta = [None] * len(self.__list)
        self.metadata = meta

    def __repr__(self):
        return "MetadataList(%s)" % self.__list

    def __len__(self):
        return len(self.__list)

    def __iter__(self):
        return self.__list.__iter__()

    def __add__(self, other):
        l = MetadataList(self, self.metadata)
        l.extend(other)
        return l

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __eq__(self, other):
        if isinstance(other, MetadataList):
            other = list(other)
        return self.__list == other

    def __getitem__(self, i):
        val = self.__list.__getitem__(i)
        if isinstance(i, slice):
            return MetadataList(val, self.metadata.__getitem__(i))
        else:
            return val

    def __setitem__(self, i, value):
        self.__list.__setitem__(i, value)
        if isinstance(value, MetadataList):
            metadata = list(value.metadata)
        else:
            metadata = [None] * len(value)
        self.metadata.__setitem__(i, metadata)

    def __delitem__(self, i):
        self.__list.__delitem__(i)
        self.metadata.__delitem__(i)

    def append(self, other, metadata=None):
        self.insert(len(self), other, metadata=metadata)

    def extend(self, other):
        self[len(self) : len(self)] = other

    def insert(self, i, item, metadata=None):
        if i >= len(self):
            i = len(self)
            e = len(self) + 1
        else:
            e = i
        self[i:e] = [item]
        self.metadata[i:e] = [metadata]

    def pop(self, i=-1):
        item = self[i]
        del self[i]
        return item

    def remove(self, item):
        del self[self.index(item)]

    def reverse(self):
        self.__list.reverse()
        self.metadata.reverse()

    def index(self, i, start=0, end=None):
        if end is None:
            return self.__list.index(i, start)
        else:
            return self.__list.index(i, start, end)

    def count(self, i):
        return self.__list.count(i)

    def get_meta_key(self, index, key, default=None):
        if not self.metadata[index]:
            return default
        return self.metadata[index].get(key, default)

    def set_meta_key(self, index, key, value):
        if not self.metadata[index]:
            self.metadata[index] = {}
        self.metadata[index][key] = value

    def del_meta_key(self, index, key):
        if not self.metadata[index]:
            raise KeyError(key)
        del self.metadata[index][key]
        if not self.metadata[index]:
            self.metadata[index] = None


class ProgressThread(GObject.GObject, threading.Thread):
    """
    A basic thread with progress updates. The thread should emit
    the progress-update signal periodically. The contents must
    be number between 0 and 100, or a tuple of (n, total) where
    n is the current step.
    """

    __gsignals__ = {
        'progress-update': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (GObject.TYPE_PYOBJECT,),
        ),
        # TODO: Check if 'stopped' is required
        'done': (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        threading.Thread.__init__(self)
        self.setDaemon(True)

    def stop(self):
        """
        Stops the thread
        """
        self.emit('done')

    def run(self):
        """
        Override and make sure that the 'progress-update'
        signal is emitted regularly with the progress
        """
        pass


class SimpleProgressThread(ProgressThread):
    """
    Simpler version of ProgressThread that uses a generator to
    manage the thread and its progress. Instead of overriding
    run, just pass a callable that returns a generator to
    the constructor.

    The callable must either yield a number between 0 and 100,
    or yield a tuple of (n, total) where n is the current step.

    ::

        def long_running_thing():
            l = len(stuff)
            try:
                for i in stuff:
                    yield (i, l)
            finally:
                # if the thread is stopped, GeneratorExit will
                # be raised the next time yield is called
                pass
    """

    def __init__(self, target, *args, **kwargs):
        ProgressThread.__init__(self)
        self.__target = (target, args, kwargs)
        self.__stop = False

    def stop(self):
        """
        Causes the thread to stop at the next yield point
        """
        self.__stop = True

    def run(self):
        """
        Runs a generator
        """
        target, args, kwargs = self.__target

        try:
            for progress in target(*args, **kwargs):
                self.emit('progress-update', progress)
                if self.__stop:
                    break
        except GeneratorExit:
            pass
        except Exception:
            logger.exception("Unhandled exception")
        finally:
            self.emit('done')


class PosetItem:
    def __init__(self, name, after, priority, value=None):
        """
        :param name: unique identifier for this item
        :type name: string
        :param after: which items this item comes after
        :type after: list of string
        :param priority: tiebreaker, higher values come later
        :type priority: int
        :param value: arbitrary data associated with the item
        """
        self.name = name
        self.after = list(after)
        self.priority = priority
        self.children = []
        self.value = value


def order_poset(items):
    """
    :param items: poset to order
    :type items: list of :class:`PosetItem`
    """
    items = {item.name: item for item in items}
    for name, item in items.items():
        for after in item.after:
            k = items.get(after)
            if k:
                k.children.append(item)
            else:
                item.after.remove(after)
    result = []
    next = [i[1] for i in items.items() if not i[1].after]
    while next:
        current = sorted((i.priority, i.name, i) for i in next)
        result.extend(i[2] for i in current)
        nextset = dict()
        for i in current:
            for c in i[2].children:
                nextset[c.name] = c
        removals = []
        for name, item in nextset.items():
            for after in item.after:
                if after in nextset:
                    removals.append(name)
                    break
        for r in removals:
            del nextset[r]
        next = list(nextset.values())
    return result


class LazyDict:
    __slots__ = ['_dict', '_funcs', 'args', '_locks']

    def __init__(self, *args):
        self.args = args
        self._dict = {}
        self._funcs = {}
        self._locks = {}

    def __setitem__(self, item, value):
        if inspect.isfunction(value):
            self._funcs[item] = value
        else:
            self._dict[item] = value

    def __getitem__(self, item):
        lock = self._locks.get(item, threading.Lock())
        with lock:
            try:
                return self._dict[item]
            except KeyError:
                self._locks[item] = lock
                val = self._funcs[item](item, *self.args)
                self._dict[item] = val
                return val

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default


class _GioFileStream:

    __slots__ = ['stream']

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.stream.close()

    def seek(self, offset, whence=os.SEEK_CUR):
        if whence == os.SEEK_CUR:
            self.stream.seek(offset, GLib.SeekType.CUR)
        elif whence == os.SEEK_SET:
            self.stream.seek(offset, GLib.SeekType.SET)
        elif whence == os.SEEK_END:
            self.stream.seek(offset, GLib.SeekType.END)
        else:
            raise IOError("Invalid whence")

    def tell(self):
        return self.stream.tell()


class GioFileInputStream(_GioFileStream):
    """
    Wrap a Gio.File so it looks like a python file object for reading.

    TODO: More complete wrapper
    """

    __slots__ = ['stream', 'gfile']

    def __init__(self, gfile):
        self.gfile = gfile
        self.stream = Gio.DataInputStream.new(gfile.read())

    def __iter__(self):
        return self

    def __next__(self):
        r = self.stream.read_line()[0]
        if not r:
            raise StopIteration()
        return r.decode('utf-8')

    def read(self, size=None):
        if size:
            return self.stream.read_bytes(size).get_data().decode('utf-8')
        else:
            return self.gfile.load_contents()[1].decode('utf-8')

    def readline(self):
        return self.stream.read_line()[0].decode('utf-8')


class GioFileOutputStream(_GioFileStream):
    """
    Wrapper around Gio.File for writing like a python file object
    """

    __slots__ = ['stream']

    def __init__(self, gfile, mode='w'):
        if mode != 'w':
            raise IOError("Not implemented")

        self.stream = gfile.replace('', False, Gio.FileCreateFlags.REPLACE_DESTINATION)

    def flush(self):
        self.stream.flush()

    def write(self, s):
        if isinstance(s, str):
            s = s.encode('utf-8', 'surrogateescape')
        return self.stream.write(s)


def subscribe_for_settings(section, options, self):
    """
    Allows you designate attributes on an object to be dynamically
    updated when a particular setting changes. If you want to be
    notified of a setting update, use a @property for the attribute.

    Only works for a options in a single section

    :param section: Settings section
    :param options: Dictionary of key: option name, value: attribute on
                    'self' to set when the setting has been updated. The
                    attribute must already have a default value in it
    :param self:    Object to set attribute values on

    :returns: A function that can be called to unsubscribe

    .. versionadded:: 3.5.0
    """

    from xl import event
    from xl import settings

    def _on_option_set(unused_name, unused_object, data):
        attrname = options.get(data)
        if attrname is not None:
            setattr(self, attrname, settings.get_option(data, getattr(self, attrname)))

    for k in options:
        if not k.startswith('%s/' % section):
            raise ValueError("Option is not part of section %s" % section)
        _on_option_set(None, None, k)

    return event.add_callback(
        _on_option_set, '%s_option_set' % section.replace('/', '_')
    )


class AsyncLoader:
    """
    Async loader based on a generator
    Threaded, load it and put it in `result_list`
    """

    def __init__(self, item_generator):
        """
        Constructs and already start processing (starts thread)
        :param item_generator: iterable
        """
        self.__end = False
        self.__result_list = []
        self.__thread = threading.Thread(target=self.run, args=(item_generator,))
        self.__thread.start()

    def run(self, item_generator):
        """
        Process items putting in `result_list`
        :param item_generator: iterable
        :return: None
        """
        for i in item_generator:
            if self.__end:
                break
            self.__result_list.append(i)
            if self.__end:
                break

    def end(self, timeout=None):
        """
        Request process ending if it doesn't occurs in timeout
        :param timeout: float representing seconds or None to wait infinitely (default)
        :return: None
        """
        self.__thread.join(timeout)
        self.__end = True

    def ended(self):
        """
        If it has ended
        :return: bool
        """
        return not self.__thread.is_alive()

    @property
    def result(self):
        """
        Gets the result
        :return: list
        """
        return self.__result_list[:]


class LowestStr(str):
    """String subclass that always sorts as the lowest value"""

    def __lt__(self, _):
        return True


class HighestStr(str):
    """String subclass that always sorts as the highest value"""

    def __lt__(self, _):
        return False


# vim: et sts=4 sw=4



db_names = {
    'dbm': 'dbm.ndbm',
    'gdbm': 'dbm.gnu',
    'dumbdbm': 'dbm.dumb',
}  # Map of old db names to new module names
available_dbs = set()  # Set of available db's (old names)

try:
    import dbm.ndbm

    available_dbs.add('dbm')
except ImportError:
    pass

try:
    import dbm.gnu

    available_dbs.add('gdbm')
except ImportError:
    pass

try:
    import dbm.dumb

    available_dbs.add('dumbdbm')
except ImportError:
    pass

import glob
import os
from os.path import basename, dirname, join
import pickle
import shutil

import pytest

from xl.common import open_shelf


@pytest.fixture(params=['dbm', 'gdbm', 'dumbdbm'])
def data(request, tmpdir):
    dbtype = request.param
    base = join(dirname(__file__), '..', '..', 'data', 'db')
    truth = {}

    # uses pickle instead of JSON because of unicode issues...
    with open(join(base, 'music.db.pickle'), 'rb') as fp:
        truth = pickle.load(fp)

    if dbtype not in available_dbs:
        pytest.skip('Module %s (%s) does not exist' % (dbtype, db_names[dbtype]))
    else:
        # copy the test data to a tempdir
        loc = str(tmpdir.mkdir(dbtype))

        for f in glob.glob(join(base, dbtype, 'music.*')):
            shutil.copyfile(f, join(loc, basename(f)))

        return truth, loc, dbtype


def test_migration(data):
    truth, loc, dbtype = data

    print(os.listdir(loc))

    try:
        db = open_shelf(join(loc, 'music.db'))
    except Exception as e:
        if dbtype == 'dbm' and getattr(e, 'args', (None,))[0] == 2:
            # on fedora it seems like dbm is linked to gdbm, and on debian based
            # systems that dbm uses a bsd implementation. Ignore these errors,
            # as (presumably) users will only try to migrate databases on systems
            # that were previously able to read the database.
            pytest.skip("Invalid dbm module")
            return
        raise

    for k, v in truth.items():
        assert k in db
        assert v == db[k]

    assert os.listdir(loc) == ['music.db']

    db.close()



# Copyright (C) 2010 Brian Parma
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
from xlgui.preferences import widgets
from xlgui import icons
from xl.nls import gettext as _

name = _('Multi-Alarm Clock')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "malrmclk.ui")

icons.MANAGER.add_icon_name_from_directory('clock', os.path.join(basedir, 'icons'))
icon = 'clock'


class FadingEnabledPreference(widgets.CheckPreference):
    name = 'plugin/multialarmclock/fading_on'
    default = True


class FadingConditional(widgets.CheckConditional):
    condition_preference_name = 'plugin/multialarmclock/fading_on'


class RestartPlaylistPreference(widgets.CheckPreference):
    name = 'plugin/multialarmclock/restart_playlist_on'
    default = False


class FadeMinVolumePreference(widgets.SpinPreference, FadingConditional):
    name = 'plugin/multialarmclock/fade_min_volume'
    default = 0

    def __init__(self, prefs, widget):
        widgets.SpinPreference.__init__(self, prefs, widget)
        FadingConditional.__init__(self)


class FadeMaxVolumePreference(widgets.SpinPreference, FadingConditional):
    name = 'plugin/multialarmclock/fade_max_volume'
    default = 100

    def __init__(self, prefs, widget):
        widgets.SpinPreference.__init__(self, prefs, widget)
        FadingConditional.__init__(self)


class FadeIncrementPreference(widgets.SpinPreference, FadingConditional):
    name = 'plugin/multialarmclock/fade_increment'
    default = 1

    def __init__(self, prefs, widget):
        widgets.SpinPreference.__init__(self, prefs, widget)
        FadingConditional.__init__(self)


class FadeTimePreference(widgets.SpinPreference, FadingConditional):
    name = 'plugin/multialarmclock/fade_time'
    default = 30

    def __init__(self, prefs, widget):
        widgets.SpinPreference.__init__(self, prefs, widget)
        FadingConditional.__init__(self)



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import GObject
from gi.repository import Gtk

from xl import trax
from xl.nls import gettext as _
from xlgui import panel
from xlgui.panel import menus
from xlgui.widgets.common import DragTreeView


class FlatPlaylistPanel(panel.Panel):
    """
    Flat playlist panel; represents a single playlist
    """

    __gsignals__ = {
        'append-items': (GObject.SignalFlags.RUN_LAST, None, (object, bool)),
        'replace-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'queue-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    ui_info = ('flatplaylist.ui', 'FlatPlaylistPanel')

    def __init__(self, parent, name, label):
        panel.Panel.__init__(self, parent, name, label)

        self.box = self.builder.get_object('FlatPlaylistPanel')
        self.model = Gtk.ListStore(int, str, object)
        self.tracks = []
        self._setup_tree()
        if not hasattr(self.parent, 'do_import'):
            self.builder.get_object("import_button").hide()
        self.menu = menus.TrackPanelMenu(self)
        self._connect_events()

    def _connect_events(self):
        self.builder.connect_signals(
            {
                'on_add_button_clicked': self._on_add_button_clicked,
                'on_import_button_clicked': self._on_import_button_clicked,
            }
        )

    def _on_add_button_clicked(self, *e):
        self.emit('append-items', self.tracks, False)

    def _on_import_button_clicked(self, *e):
        tracks = self.tree.get_selected_tracks()
        if len(tracks) == 0:  # nothing selected, do everything
            tracks = self.tracks
        self.parent.do_import(tracks)

    def _setup_tree(self):
        self.tree = FlatPlaylistDragTreeView(self, False, True)
        selection = self.tree.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)

        self.tree.set_headers_visible(True)
        self.tree.set_model(self.model)
        self.scroll = Gtk.ScrolledWindow()
        self.scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.scroll.add(self.tree)
        self.scroll.set_shadow_type(Gtk.ShadowType.IN)
        self.box.pack_start(self.scroll, True, True, 0)

        text = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn(_('#'))
        col.pack_start(text, False)
        col.set_attributes(text, text=0)
        col.set_fixed_width(50)
        col.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self.tree.append_column(col)

        text = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn(_('Title'))
        col.pack_start(text, True)
        col.set_attributes(text, text=1)
        col.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        col.set_cell_data_func(text, self._title_data_func)
        self.tree.append_column(col)
        self.box.show_all()

    def _title_data_func(self, col, cell, model, iter, _unused):
        if not model.iter_is_valid(iter):
            return
        item = model.get_value(iter, 2)
        cell.set_property('text', item.get_tag_display("title"))

    def set_playlist(self, playlist):
        self.model.clear()

        self.tracks = tracks = list(playlist)
        for i, track in enumerate(tracks):
            self.model.append([i + 1, track.get_tag_display("title"), track])

    def drag_data_received(self, *e):
        """
        stub
        """
        pass

    def drag_data_delete(self, *e):
        """
        stub
        """
        pass

    def drag_get_data(self, treeview, context, selection, target_id, etime):
        """
        Called when a drag source wants data for this drag operation
        """
        tracks = self.tree.get_selected_tracks()
        if not tracks:
            return
        for track in tracks:
            DragTreeView.dragged_data[track.get_loc_for_io()] = track
        uris = trax.util.get_uris_from_tracks(tracks)
        selection.set_uris(uris)


class FlatPlaylistDragTreeView(DragTreeView):
    """
    Custom DragTreeView to retrieve data from playlists
    """

    def get_selected_tracks_count(self):
        """
        Returns the count of selected tracks
        """
        return self.get_selection().count_selected_rows()

    def get_selected_tracks(self):
        """
        Returns the currently selected tracks
        """
        (model, paths) = self.get_selection().get_selected_rows()
        tracks = []

        for path in paths:
            iter = model.get_iter(path)
            track = model.get_value(iter, 2)
            tracks.append(track)

        return tracks



# Copyright (C) 2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from xl import xdg
from xl.nls import gettext as _
from xlgui.preferences import widgets

name = _('Collection')
icon = 'folder-music'
ui = xdg.get_data_path('ui', 'preferences', 'collection.ui')


def _get_default_strip_list():
    return []
    # FIXME:  currently, this is broken by the backend not also having access
    # to the default set here, so we will just NOT set one.

    # TRANSLATORS: Grammatical articles that are ignored while sorting the
    # collection panel. For example, in French locales this could be
    # the space-separated list "l' la le les".
    # If this practice is not common in your locale, simply
    # translate this to string with single space.
    default_strip_list = _("the")
    return [v.lower() for v in default_strip_list.split(' ') if v]


class CollectionStripArtistPreference(widgets.ListPreference):
    default = _get_default_strip_list()
    name = 'collection/strip_list'

    def __init__(self, preferences, widget):
        widgets.ListPreference.__init__(self, preferences, widget)
        self.widget.connect('populate-popup', self._populate_popup_cb)

    def _get_value(self):
        """
        Get the value, overrides the base class function
        because we don't need shlex parsing. We actually
        want values like "l'" here.
        """
        values = [v.lower() for v in self.widget.get_text().split(' ') if v]
        return values

    def _populate_popup_cb(self, entry, menu):
        from gi.repository import Gtk

        entry = Gtk.MenuItem.new_with_mnemonic(_('Reset to _Defaults'))
        entry.connect('activate', self._reset_to_defaults_cb)
        entry.show()

        sep = Gtk.SeparatorMenuItem()
        sep.show()

        menu.attach(entry, 0, 1, 0, 1)
        menu.attach(sep, 0, 1, 1, 2)

    def _reset_to_defaults_cb(self, item):
        self.widget.set_text(' '.join(_get_default_strip_list()))


class FileBasedCompilationsPreference(widgets.CheckPreference):
    default = True
    name = 'collection/file_based_compilations'


# vim:ts=4 et sw=4



from gi.repository import GLib
import pytest

from xl.player.track_fader import TrackFader, FadeState

NoFade = FadeState.NoFade
FadingIn = FadeState.FadingIn
Normal = FadeState.Normal
FadingOut = FadeState.FadingOut


class FakeStream:
    def __init__(self):
        self.reset()

    def reset(self):
        self.position = 0
        self.volume = 42
        self.fadeout_begin = False
        self.stopped = False

    def get_position(self):
        return self.position

    def set_volume(self, value):
        self.volume = int(value * 100)

    def stop(self):
        self.stopped = True

    def on_fade_out(self):
        self.fadeout_begin = True


class FakeTrack:
    def __init__(self, start_off, stop_off, tracklen):
        self.tags = {
            '__startoffset': start_off,
            '__stopoffset': stop_off,
            '__length': tracklen,
        }

    def get_tag_raw(self, t):
        return self.tags[t]


# TODO: monkeypatch instead
timeout_args = [()]


def glib_timeout_add(*args):
    timeout_args[0] = args[2:]
    return args[1]


def glib_source_remove(src_id):
    pass


GLib.timeout_add = glib_timeout_add
GLib.source_remove = glib_source_remove

TmSt = 1
TmEx = 2


# Test data:
#   Position, Volume, State, TmSt/TmEx/None, [call, [arg1...]]

# fmt: off
@pytest.mark.parametrize('test', [

    # Test don't manage the volume
    [
        (0, 100, NoFade, None, 'play', None, None, None, None),
        (1, 100, NoFade, None, 'pause'),
        (2, 100, NoFade, None, 'unpause'),
        (3, 100, NoFade, None, 'seek', 4),
        (4, 100, NoFade, None, 'stop'),
        (5, 100, NoFade, None),
    ],

    # Test fading in
    [
        (0, 0,  FadingIn, TmEx, 'play', 0, 2, None, None),
        (1, 50, FadingIn, TmEx, 'execute'),
        (3, 100, NoFade,  None, 'execute'),
        (4, 100, NoFade,  None),
        (5, 100, NoFade,  None, 'stop'),
        (6, 100, NoFade,  None),
    ],

    # Test fading in: pause in middle
    [
        (0, 0,  FadingIn, TmEx, 'play', 0, 2, None, None),
        (1, 50, FadingIn, TmEx, 'execute'),
        (1, 50, FadingIn, None, 'pause'),
        (1, 50, FadingIn, TmEx, 'unpause'),
        (1, 50, FadingIn, TmEx, 'execute'),
        (3, 100, NoFade,  None, 'execute'),
        (4, 100, NoFade,  None),
        (5, 100, NoFade,  None, 'stop'),
        (6, 100, NoFade,  None),
    ],

    # Test fading in past the fade point
    [
        (3, 100, NoFade, None, 'play', 0, 2, None, None),
        (4, 100, NoFade, None),
        (5, 100, NoFade, None, 'stop'),
        (6, 100, NoFade, None),
    ],

    # Test fading out
    [
        (3, 100, Normal,    TmSt, 'play', None, None, 4, 6),
        (4, 100, FadingOut, TmEx, 'start'),
        (5, 50,  FadingOut, TmEx, 'execute'),
        (6, 0,   FadingOut, TmEx, 'execute'),
        (6.1, 0, NoFade,    None, 'execute'),
        (7,   0, NoFade,    None),
    ],

    # Test all of them
    [
        (0, 0,  FadingIn,   TmEx, 'play', 0, 2, 4, 6),
        (1, 50, FadingIn,   TmEx, 'execute'),
        (3, 100, Normal,    TmSt, 'execute'),
        (4, 100, FadingOut, TmEx, 'start'),
        (5, 50,  FadingOut, TmEx, 'execute'),
        (6, 0,   FadingOut, TmEx, 'execute'),
        (6.1, 0, NoFade,    None, 'execute'),
        (7,   0, NoFade,    None),
    ],

    # Test fading in with startoffset
    # [
    #     (0, 0,  FadingIn,  TmEx, 'play', 60, 62, 64, 66),
    #     (0, 0,  FadingIn,  TmEx, 'seek', 60),
    #     (61, 50,  FadingIn,  TmEx, 'execute'),
    # ],
])
# fmt: on
def test_fader(test):

    # Test fade_out_on_play

    # Test setup_track

    # Test setup_track is_update=True

    # Test unexpected fading out

    check_fader(test)


def check_fader(test):
    stream = FakeStream()
    fader = TrackFader(stream, stream.on_fade_out, 'test')

    for data in test:
        print(data)
        now = data[0]
        stream.position = int(now * TrackFader.SECOND)
        print(stream.position)
        volume = data[1]
        state = data[2]
        timer_id = data[3]

        if len(data) > 4:
            action = data[4]
            args = data[5:] if len(data) > 5 else ()

            if action == 'start':
                action = '_on_fade_start'
            elif action == 'execute':
                action = '_execute_fade'
                args = timeout_args[0]
                fader.now = now - 0.010

            # Call the function
            getattr(fader, action)(*args)

        # Check to see if timer id exists
        if timer_id is None:
            assert fader.timer_id is None
        elif timer_id == TmSt:
            assert fader.timer_id == fader._on_fade_start
        elif timer_id == TmEx:
            assert fader.timer_id == fader._execute_fade
        else:
            assert False

        assert fader.state == state
        assert stream.volume == volume


def test_calculate_fades():
    fader = TrackFader(None, None, None)

    # fin, fout, start_off, stop_off, tracklen;
    # start, start+fade, end-fade, end
    calcs = [
        # fmt: off
        
        # one is zero/none
        (0, 4, 0, 0, 10,        0, 0, 6, 10),
        (None, 4, 0, 0, 10,     0, 0, 6, 10),

        # other is zero/none
        (4, 0, 0, 0, 10,        0, 4, 10, 10),
        (4, None, 0, 0, 10,     0, 4, 10, 10),

        # both are equal
        (4, 4, 0, 0, 10,        0, 4, 6, 10),

        # both are none
        (0, 0, 0, 0, 10,        0, 0, 10, 10),
        (None, None, 0, 0, 10,  0, 0, 10, 10),

        # Bigger than playlen: all three cases
        (0, 4, 0, 0, 2,         0, 0, 0, 2),
        (4, 0, 0, 0, 2,         0, 2, 2, 2),
        (4, 4, 0, 0, 2,         0, 1, 1, 2),

        # With start offset
        (4, 4, 1, 0, 10,        1, 5, 6, 10),

        # With stop offset
        (4, 4, 0, 9, 10,        0, 4, 5, 9),

        # With both
        (2, 2, 1, 9, 10,        1, 3, 7, 9),

        # With both, constrained
        (4, 4, 4, 8, 10,        4, 6, 6, 8),
        (2, 4, 4, 7, 10,        4, 5, 5, 7),
        (4, 2, 4, 7, 10,        4, 6, 6, 7),
        # fmt: on
    ]

    i = 0
    for fin, fout, start, stop, tlen, t0, t1, t2, t3 in calcs:
        print(
            '%2d: Fade In: %s; Fade Out: %s; start: %s; stop: %s; Len: %s'
            % (i, fin, fout, start, stop, tlen)
        )
        track = FakeTrack(start, stop, tlen)
        assert fader.calculate_fades(track, fin, fout) == (t0, t1, t2, t3)
        i += 1



# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from xl import providers, event, settings
from xl.player.gst.gst_utils import ElementBin

from gi.repository import Gst

try:
    from . import replaygainprefs

    def get_preferences_pane():
        return replaygainprefs


except Exception:  # fail gracefully if we cant set up the UI
    pass

NEEDED_ELEMS = ["rgvolume", "rglimiter"]


def enable(exaile):
    for elem in NEEDED_ELEMS:
        if not Gst.ElementFactory.find(elem):
            raise ImportError("Needed gstreamer element %s missing." % elem)
    providers.register("gst_audio_filter", ReplaygainVolume)
    providers.register("gst_audio_filter", ReplaygainLimiter)


def disable(exaile):
    providers.unregister("gst_audio_filter", ReplaygainVolume)
    providers.unregister("gst_audio_filter", ReplaygainLimiter)


class ReplaygainVolume(ElementBin):
    """
    Handles replaygain volume adjustment and pre-amp.

    Placed at 20 in the pipeline, since most elements should do their
    processing after it.
    """

    index = 20
    name = "rgvolume"

    def __init__(self):
        ElementBin.__init__(self, name=self.name)
        self.audioconvert = Gst.ElementFactory.make("audioconvert", None)
        self.elements[40] = self.audioconvert
        self.rgvol = Gst.ElementFactory.make("rgvolume", None)
        self.elements[50] = self.rgvol
        self.setup_elements()

        event.add_ui_callback(self._on_option_set, "replaygain_option_set")

        # load settings
        for x in ("album-mode", "pre-amp", "fallback-gain"):
            self._on_option_set("replaygain_option_set", None, "replaygain/%s" % x)

    def _on_option_set(self, name, object, data):
        if data == "replaygain/album-mode":
            self.rgvol.set_property(
                "album-mode", settings.get_option("replaygain/album-mode", True)
            )
        elif data == "replaygain/pre-amp":
            self.rgvol.set_property(
                "pre-amp", settings.get_option("replaygain/pre-amp", 0)
            )
        elif data == "replaygain/fallback-gain":
            self.rgvol.set_property(
                "fallback-gain", settings.get_option("replaygain/fallback-gain", 0)
            )


class ReplaygainLimiter(ElementBin):
    """
    Implements clipping protection.

    Placed at 80 in the pipeline so that other elements can come
    before it if necessary.
    """

    index = 80
    name = "rglimiter"

    def __init__(self):
        ElementBin.__init__(self, name=self.name)
        self.rglimit = Gst.ElementFactory.make("rglimiter", None)
        self.elements[50] = self.rglimit
        self.audioconvert = Gst.ElementFactory.make("audioconvert", None)
        self.elements[60] = self.audioconvert
        self.setup_elements()

        event.add_ui_callback(self._on_option_set, "replaygain_option_set")
        self._on_option_set(
            "replaygain_option_set", None, "replaygain/clipping-protection"
        )

    def _on_option_set(self, name, object, data):
        if data == "replaygain/clipping-protection":
            self.rglimit.set_property(
                "enabled", settings.get_option("replaygain/clipping-protection", True)
            )



# Copyright (C) 2011 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk
from gi.repository import Pango

from xl import event, providers, player, settings

from xl.nls import gettext as _

from xlgui.widgets import menu, dialogs

from . import gt_prefs
from . import gt_widgets
from .gt_common import (
    migrate_settings,
    get_group_categories,
    set_group_categories,
    get_track_groups,
    set_track_groups,
    create_all_search_playlist,
    tagname_option,
    create_custom_search_playlist,
)
from . import gt_export
from . import gt_import
from . import gt_mass


class GroupTaggerPlugin:
    '''Implements logic for plugin'''

    def get_preferences_pane(self):
        return gt_prefs

    def enable(self, exaile):
        self.exaile = exaile

    def on_gui_loaded(self):

        self.track = None
        self.tag_dialog = None

        migrate_settings()

        self.panel = gt_widgets.GroupTaggerPanel(self.exaile)
        self.panel.show_all()
        self.setup_panel_font(False)

        self.panel.tagger.view.connect('category-changed', self.on_category_change)
        self.panel.tagger.view.connect('category-edited', self.on_category_edited)
        self.panel.tagger.view.connect('group-changed', self.on_group_change)

        # add to exaile's panel interface
        providers.register('main-panel', self.panel)

        # ok, register for some events
        event.add_ui_callback(self.on_playback_track_start, 'playback_track_start')
        event.add_ui_callback(
            self.on_playlist_cursor_changed, 'playlist_cursor_changed'
        )
        event.add_ui_callback(
            self.on_plugin_options_set, 'plugin_grouptagger_option_set'
        )

        # add our own submenu for functionality
        tools_submenu = menu.Menu(None, context_func=lambda p: self.exaile)

        tools_submenu.add_item(
            menu.simple_menu_item(
                'gt_get_tags',
                [],
                _('_Get all tags from collection'),
                callback=self.on_get_tags_menu,
            )
        )

        tools_submenu.add_item(
            menu.simple_menu_item(
                'gt_import',
                [],
                _('_Import tags from directory'),
                callback=self.on_import_tags,
            )
        )

        tools_submenu.add_item(
            menu.simple_menu_item(
                'gt_rename',
                [],
                _('_Mass rename/delete tags'),
                callback=self.on_mass_rename,
            )
        )

        tools_submenu.add_item(
            menu.simple_menu_item(
                'gt_export',
                [],
                _('E_xport collecton tags to JSON'),
                callback=self.on_export_tags,
            )
        )

        # group them together to make it not too long
        self.tools_menuitem = menu.simple_menu_item(
            'grouptagger', ['plugin-sep'], _('_GroupTagger'), submenu=tools_submenu
        )
        providers.register('menubar-tools-menu', self.tools_menuitem)

        # playlist context menu items
        self.provider_items = []
        track_subitem = menu.Menu(None, inherit_context=True)

        track_subitem.add_item(
            menu.simple_menu_item(
                'gt_search_all',
                [],
                _('Show tracks with all tags'),
                callback=self.on_playlist_context_select_all_menu,
                callback_args=[self.exaile],
            )
        )

        track_subitem.add_item(
            menu.simple_menu_item(
                'gt_search_custom',
                ['gt_search_all'],
                _('Show tracks with tags (custom)'),
                callback=self.on_playlist_context_select_custom_menu,
                callback_args=[self.exaile],
            )
        )

        tag_cond_fn = lambda n, p, c: c['selection-count'] > 1

        track_subitem.add_item(
            menu.simple_menu_item(
                'gt_tag_add_multi',
                ['gt_search_custom'],
                _('Add tags to all'),
                callback=self.on_add_tags,
                condition_fn=tag_cond_fn,
                callback_args=[self.exaile],
            )
        )

        track_subitem.add_item(
            menu.simple_menu_item(
                'gt_tag_rm_multi',
                ['gt_tag_add_multi'],
                _('Remove tags from all'),
                callback=self.on_rm_tags,
                condition_fn=tag_cond_fn,
                callback_args=[self.exaile],
            )
        )

        self.provider_items.append(
            menu.simple_menu_item(
                'grouptagger', ['rating'], _('GroupTagger'), submenu=track_subitem
            )
        )

        for item in self.provider_items:
            providers.register('playlist-context-menu', item)
            # Hm, doesn't work..
            # providers.register('track-panel-menu', item)

        # trigger start event if exaile is currently playing something
        if player.PLAYER.is_playing():
            self.set_display_track(player.PLAYER.current)
        else:
            self.panel.tagger.set_categories([], get_group_categories())

    def disable(self, exaile):
        '''Called when the plugin is disabled'''

        if self.tools_menuitem:
            providers.unregister('menubar-tools-menu', self.tools_menuitem)
            for item in self.provider_items:
                providers.unregister('playlist-context-menu', item)
                providers.unregister('track-panel-menu', item)

            self.tools_menuitem = None
            self.provider_items = []

        if self.tag_dialog:
            self.tag_dialog.destroy()
            self.tag_dialog = None

        # de-register the exaile events
        event.remove_callback(self.on_playback_track_start, 'playback_track_start')
        event.remove_callback(
            self.on_playlist_cursor_changed, 'playlist_cursor_changed'
        )
        event.remove_callback(
            self.on_plugin_options_set, 'plugin_grouptagger_option_set'
        )

        providers.unregister('main-panel', self.panel)

    def setup_panel_font(self, always_set):
        font = settings.get_option('plugin/grouptagger/panel_font', None)
        if font is None:
            if not always_set:
                return
            font = gt_prefs._get_system_default_font()
        else:
            font = Pango.FontDescription(font)

        self.panel.tagger.set_font(font)

    #
    # Menu callbacks
    #

    def on_export_tags(self, widget, name, parent, exaile):
        gt_export.export_tags(exaile)

    def on_get_tags_menu(self, widget, name, parent, exaile):

        if self.tag_dialog is None:
            self.tag_dialog = gt_widgets.AllTagsDialog(
                exaile, self.panel.tagger.add_groups
            )
            self.tag_dialog.connect(
                'delete-event', self.on_get_tags_menu_window_deleted
            )

        self.tag_dialog.show_all()

    def on_get_tags_menu_window_deleted(self, *args):
        self.tag_dialog = None

    def on_import_tags(self, widget, name, parent, exaile):
        gt_import.import_tags(exaile)

    def on_mass_rename(self, widget, name, parent, exaile):
        gt_mass.mass_rename(exaile)

    def _add_rm_multi_tags(self, add, context, exaile):
        '''Add or remove tags from multiple tracks'''
        tracks = context['selected-tracks']

        dialog = gt_widgets.GroupTaggerAddRemoveDialog(add, tracks, exaile)
        if add:
            dialog.tagger.set_categories([], get_group_categories())
        else:
            groups = set()
            for track in tracks:
                groups |= get_track_groups(track)
            dialog.tagger.add_groups(groups)

        # TODO: something more dynamic
        dialog.set_size_request(250, 500)

        retval = dialog.run()

        groups = {}

        if retval == Gtk.ResponseType.APPLY:
            groups = set(dialog.get_active())

        dialog.destroy()

        if len(groups) > 0:
            for track in tracks:
                existing = get_track_groups(track)
                if add:
                    set_track_groups(track, existing | groups)
                else:
                    set_track_groups(track, existing - groups)

    def on_add_tags(self, widget, name, parent, context, exaile):
        self._add_rm_multi_tags(True, context, exaile)

    def on_rm_tags(self, widget, name, parent, context, exaile):
        self._add_rm_multi_tags(False, context, exaile)

    #
    # Exaile events
    #

    def on_playback_track_start(self, type, player, track):
        '''Called when a new track starts'''
        self.set_display_track(track)

    def on_playlist_context_select_all_menu(
        self, menu, display_name, playlist_view, context, exaile
    ):
        '''Called when 'Select tracks with same tags' is selected'''
        tracks = context['selected-tracks']
        groups = set()

        for track in tracks:
            groups |= get_track_groups(track)

        if len(groups) > 0:
            create_all_search_playlist(groups, exaile)
        else:
            dialogs.error(None, _('No categorization tags found in selected tracks'))

    def on_playlist_context_select_custom_menu(
        self, menu, display_name, playlist_view, context, exaile
    ):
        '''Called when 'select tracks with similar tags (custom)' is selected'''
        tracks = context['selected-tracks']
        groups = set()

        for track in tracks:
            groups |= get_track_groups(track)

        if len(groups) > 0:
            create_custom_search_playlist(groups, exaile)
        else:
            dialogs.error(None, _('No categorization tags found in selected tracks'))

    def on_playlist_cursor_changed(self, type, playlist_view, context):
        '''Called when an item in a playlist is selected'''

        # TODO: Allow multiple tracks
        tracks = context['selected-tracks']
        if len(tracks) == 1:
            self.set_display_track(tracks[0])

    def set_display_track(self, track, force_update=False):
        '''Updates the display with the tags/info for a particular track'''

        if self.track == track and not force_update:
            return

        self.track = track

        # get the groups as a set
        track_groups = get_track_groups(track)

        # set them
        self.panel.tagger.view.show_click_column()
        self.panel.tagger.set_categories(track_groups, get_group_categories())
        self.panel.tagger.set_track_info(track)

    #
    # Widget events
    #

    def on_category_change(self, view, action, category):
        '''Called when a category has something happen to it'''

        categories = get_group_categories()

        if action == gt_widgets.category_change.added:
            categories.setdefault(category, [True, []])

        elif action == gt_widgets.category_change.deleted:
            del categories[category]

        elif action == gt_widgets.category_change.collapsed:
            categories[category][0] = False

        elif action == gt_widgets.category_change.expanded:
            categories[category][0] = True

        elif action == gt_widgets.category_change.updated:
            v = categories.setdefault(category, [True, []])
            v[1] = view.get_model().get_category_groups(category)

        set_group_categories(categories)

    def on_category_edited(self, view, old_category, new_category):
        '''Called when a category name is edited'''

        categories = get_group_categories()
        categories[new_category] = categories.pop(old_category)
        set_group_categories(categories)

    def on_group_change(self, view, action, value):
        '''Called when a group is added/deleted/updated on the widget'''

        if self.track is not None:
            groups = view.get_model().iter_active()
            if not set_track_groups(self.track, groups):
                self.set_display_track(self.track, force_update=True)

    def on_plugin_options_set(self, evtype, settings, option):
        '''Handles option changes'''
        if option == 'plugin/grouptagger/panel_font':
            self.setup_panel_font(True)
        elif option == tagname_option:
            if self.track is not None:
                self.set_display_track(self.track, True)


plugin_class = GroupTaggerPlugin



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from xlgui.preferences import widgets
from xl import common, xdg
from xl.nls import gettext as _

# TODO: If we ever add another engine, need to make sure that
#       gstreamer-specific stuff doesn't accidentally get loaded
from xl.player.gst.sink import get_devices, SINK_PRESETS

name = _('Playback')
icon = 'media-playback-start'
ui = xdg.get_data_path('ui', 'preferences', 'playback.ui')


class EnginePreference(widgets.ComboPreference):
    default = "gstreamer"
    name = 'player/engine'
    restart_required = True


class AudioSinkPreference(widgets.ComboPreference):
    default = "auto"
    name = 'player/audiosink'

    def __init__(self, preferences, widget):
        widgets.ComboPreference.__init__(self, preferences, widget)
        model = self.widget.get_model()

        # always list auto first, custom last
        def keyfunc(item):
            name = item[0]
            if name == 'auto':
                return common.LowestStr(name)
            elif name == 'custom':
                return common.HighestStr(name)
            else:
                return name

        for name, preset in sorted(SINK_PRESETS.items(), key=keyfunc):
            model.append((name, preset['name']))
        self._set_value()


class CustomAudioSinkPreference(widgets.Preference, widgets.Conditional):
    default = ""
    name = "player/custom_sink_pipe"
    condition_preference_name = 'player/audiosink'

    def __init__(self, preferences, widget):
        widgets.Preference.__init__(self, preferences, widget)
        widgets.Conditional.__init__(self)

    def on_check_condition(self):
        return self.get_condition_value() == 'custom'

    def on_condition_met(self):
        self.show_widget()

    def on_condition_failed(self):
        self.hide_widget()


class SelectDeviceForSinkPreference(widgets.ComboPreference, widgets.Conditional):
    default = 'auto'
    name = "player/audiosink_device"
    condition_preference_name = 'player/audiosink'

    def __init__(self, preferences, widget):
        self.is_enabled = False
        widgets.ComboPreference.__init__(self, preferences, widget)
        widgets.Conditional.__init__(self)

    def on_check_condition(self):
        return self.get_condition_value() == 'auto'

    def on_condition_met(self):

        # disable because the clear() causes a settings write
        self.is_enabled = False

        model = self.widget.get_model()
        if model is None:
            return

        model.clear()

        for device_name, device_id, _create_audiosink_cb in get_devices():
            model.append((device_id, device_name))

        self.is_enabled = True
        self._set_value()

        self.show_widget()
        self.set_widget_sensitive(True)

    def on_condition_failed(self):
        if self.get_condition_value() == 'custom':
            self.hide_widget()
        else:
            self.show_widget()
            self.set_widget_sensitive(False)
        self.is_enabled = False
        model = self.widget.get_model()
        if model:
            model.clear()

    def done(self):
        return self.is_enabled

    def _get_value(self):
        if self.is_enabled:
            return widgets.ComboPreference._get_value(self)
        return ''


class ResumePreference(widgets.CheckPreference):
    default = True
    name = 'player/resume_playback'


class PausedPreference(widgets.CheckPreference):
    default = False
    name = 'player/resume_paused'


class EnqueueBeginsPlayback(widgets.CheckPreference):
    default = True
    name = 'queue/enqueue_begins_playback'


class RemoveQueuedItemWhenPlayed(widgets.CheckPreference):
    default = True
    name = 'queue/remove_item_when_played'


class DisableNewTrackWhenPlaying(widgets.CheckPreference):
    default = False
    name = 'queue/disable_new_track_when_playing'


class GaplessPlayback(widgets.CheckPreference):
    default = True
    name = 'player/gapless_playback'


class EngineConditional(widgets.Conditional):
    """
    True if the specified engine is selected
    """

    condition_preference_name = 'player/engine'
    conditional_engine = ''

    def on_check_condition(self):
        if self.get_condition_value() == self.conditional_engine:
            return True

        return False


class AutoAdvancePlayer(widgets.CheckPreference, EngineConditional):
    default = True
    name = 'player/auto_advance'
    conditional_engine = 'gstreamer'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        EngineConditional.__init__(self)


class AutoAdvanceDelay(widgets.SpinPreference, widgets.MultiConditional):
    default = 0
    name = "player/auto_advance_delay"
    condition_preference_names = ['player/auto_advance', 'player/engine']

    def __init__(self, preferences, widget):
        widgets.SpinPreference.__init__(self, preferences, widget)
        widgets.MultiConditional.__init__(self)

    def on_check_condition(self):
        if not self.condition_widgets['player/auto_advance'].get_active():
            return False

        if self.get_condition_value('player/engine') == 'gstreamer':
            return True

        return False


class UserFadeTogglePreference(widgets.CheckPreference, EngineConditional):
    default = False
    name = 'player/user_fade_enabled'
    conditional_engine = 'gstreamer'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        EngineConditional.__init__(self)


class UserFadeDurationPreference(widgets.SpinPreference, EngineConditional):
    default = 1000
    name = 'player/user_fade'
    conditional_engine = 'gstreamer'

    def __init__(self, preferences, widget):
        widgets.SpinPreference.__init__(self, preferences, widget)
        EngineConditional.__init__(self)


class CrossfadingPreference(widgets.CheckPreference, EngineConditional):
    default = False
    name = 'player/crossfading'
    conditional_engine = 'gstreamer'

    def __init__(self, preferences, widget):
        widgets.CheckPreference.__init__(self, preferences, widget)
        EngineConditional.__init__(self)


class CrossfadeDurationPreference(widgets.SpinPreference, EngineConditional):
    default = 1000
    name = 'player/crossfade_duration'
    conditional_engine = 'gstreamer'

    def __init__(self, preferences, widget):
        widgets.SpinPreference.__init__(self, preferences, widget)
        EngineConditional.__init__(self)


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from xl.metadata._base import BaseFormat, CoverImage
from mutagen import id3


class ID3Format(BaseFormat):
    MutagenType = id3.ID3
    tag_mapping = {
        "originalalbum": "TOAL",
        "lyricist": "TEXT",
        "part": "TSST",
        "website": "WOAR",
        "cover": "APIC",
        "originalartist": "TOPE",
        "author": "TOLY",
        "originaldate": "TDOR",
        "date": "TDRC",
        "arranger": "TPE4",
        "conductor": "TPE3",
        "albumartist": "TPE2",
        "artist": "TPE1",
        "album": "TALB",
        "copyright": "TCOP",
        "lyrics": "USLT",
        "tracknumber": "TRCK",
        "version": "TIT3",
        "title": "TIT2",
        "isrc": "TSRC",
        "genre": "TCON",
        "composer": "TCOM",
        "encodedby": "TENC",
        "organization": "TPUB",
        "discnumber": "TPOS",
        "bpm": "TBPM",
        "grouping": "TIT1",
        "comment": "COMM",
        'language': "TLAN",
    }
    writable = True
    others = False  # make this true once custom tag support actually works

    def get_keys_disk(self):
        keys = []
        for v in self._get_raw().values():
            # Have to use this because some ID3 tags have a colon
            # in them, and so self._get_raw().keys() doesn't return
            # the expected value
            k = v.FrameID
            if k in self._reverse_mapping:
                keys.append(self._reverse_mapping[k])
            else:
                keys.append(k)
        return keys

    def _get_tag(self, raw, t):
        if not raw.tags:
            return []
        if t not in self.tag_mapping.values():
            t = "TXXX:" + t
        field = raw.tags.getall(t)
        if len(field) <= 0:
            return []
        ret = []
        if t in ('TDRC', 'TDOR'):  # values are ID3TimeStamps, need str conversion
            for value in field:
                ret.extend([str(x) for x in value.text])
        elif t == 'USLT':  # Lyrics are stored in a single str object
            for value in field:
                ret.append(value.text)
        elif (
            t == 'WOAR'
        ):  # URLs are stored in url field instead of text field (as a single str object)
            for value in field:
                ret.append(value.url.replace('\n', '').replace('\r', ''))
        elif t == 'APIC':
            ret = [
                CoverImage(type=f.type, desc=f.desc, mime=f.mime, data=f.data)
                for f in field
            ]
        elif t == 'COMM':  # Newlines within comments are allowed, keep them
            for item in field:
                ret.extend([value for value in item.text])
        else:
            for value in field:
                try:
                    ret.extend(
                        [x.replace('\n', '').replace('\r', '') for x in value.text]
                    )
                except Exception:
                    pass
        return ret

    def _set_tag(self, raw, tag, data):
        if tag not in self.tag_mapping.values():
            tag = "TXXX:" + tag

        if raw.tags is not None:
            raw.tags.delall(tag)

        # FIXME: Properly set and retrieve multiple values
        if tag == 'USLT':
            data = data[0]

        if tag == 'APIC':
            frames = [
                id3.Frames[tag](
                    encoding=3,
                    mime=info.mime,
                    type=info.type,
                    desc=info.desc,
                    data=info.data,
                )
                for info in data
            ]
        elif tag == 'COMM':
            frames = [
                id3.COMM(encoding=3, text=d, desc='', lang='\x00\x00\x00') for d in data
            ]
        elif tag == 'WOAR':
            frames = [id3.WOAR(encoding=3, url=d) for d in data]
        else:
            frames = [id3.Frames[tag](encoding=3, text=data)]

        if raw.tags is not None:
            for frame in frames:
                raw.tags.add(frame)

    def _del_tag(self, raw, tag):
        if tag not in self.tag_mapping.values():
            tag = "TXXX:" + tag
        if raw.tags is not None:
            raw.tags.delall(tag)


# vim: et sts=4 sw=4



#!/usr/bin/env python3
'''
Provides IPython console widget.

@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made
available under the terms of the BSD which accompanies this distribution, and
is available at U{https://www.opensource.org/licenses/bsd-license.php}
'''

# Taken from [1] (rev 36e2742, 2019-07-27) with slight modifications.
# The exact license [2] is reproduced below (3-clause BSD).
#
# [1] https://git.gnome.org/browse/accerciser/tree/plugins/ipython_view.py
# [2] https://git.gnome.org/browse/accerciser/tree/COPYING

# Copyright (c) 2005, 2007 IBM Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#     * Neither the name of the IBM Corporation nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import re
import sys
import os
import subprocess

from io import StringIO
from functools import reduce

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Pango

import IPython


IPYTHON_VERSION = int(IPython.__version__.split('.', 1)[0])


class IterableIPShell:
    """
    Create an IPython instance. Does not start a blocking event loop,
    instead allow single iterations. This allows embedding in GTK+
    without blockage.

    @ivar IP: IPython instance.
    @type IP: IPython.iplib.InteractiveShell
    @ivar iter_more: Indicates if the line executed was a complete command,
    or we should wait for more.
    @type iter_more: integer
    @ivar history_level: The place in history where we currently are
    when pressing up/down.
    @type history_level: integer
    @ivar complete_sep: Separation delimiters for completion function.
    @type complete_sep: _sre.SRE_Pattern
    """

    def __init__(self, user_ns=None, cin=None, cout=None, cerr=None, input_func=None):
        """

        @param argv: Command line options for IPython
        @type argv: list
        @param user_ns: User namespace.
        @type user_ns: dictionary
        @param user_global_ns: User global namespace.
        @type user_global_ns: dictionary.
        @param cin: Console standard input.
        @type cin: IO stream
        @param cout: Console standard output.
        @type cout: IO stream
        @param cerr: Console standard error.
        @type cerr: IO stream
        @param input_func: Replacement for builtin raw_input()
        @type input_func: function
        """

        self.prompt = None

        ip_io = IPython.utils.io
        if input_func:
            IPython.terminal.interactiveshell.raw_input_original = input_func
        if cin:
            ip_io.stdin = ip_io.IOStream(cin)
        if cout:
            ip_io.stdout = ip_io.IOStream(cout)
        if cerr:
            ip_io.stderr = ip_io.IOStream(cerr)

        # This is to get rid of the blockage that occurs during
        # IPython.Shell.InteractiveShell.user_setup()

        ip_io.raw_input = lambda x: None

        os.environ['TERM'] = 'dumb'
        excepthook = sys.excepthook

        if IPYTHON_VERSION >= 4:
            from traitlets.config.loader import Config
        else:
            from IPython.config.loader import Config

        cfg = Config()
        cfg.InteractiveShell.colors = "Linux"
        cfg.Completer.use_jedi = False

        # InteractiveShell's __init__ overwrites ip_io.stdout,ip_io.stderr with
        # sys.stdout, sys.stderr, this makes sure they are right
        #
        old_stdout, old_stderr = sys.stdout, sys.stderr
        sys.stdout, sys.stderr = ip_io.stdout.stream, ip_io.stderr.stream

        # InteractiveShell inherits from SingletonConfigurable, so use instance()
        #
        self.IP = IPython.terminal.embed.InteractiveShellEmbed.instance(
            config=cfg, user_ns=user_ns
        )

        sys.stdout, sys.stderr = old_stdout, old_stderr

        self.IP.system = lambda cmd: self.shell(
            self.IP.var_expand(cmd), header='IPython system call: '
        )
        # local_ns=user_ns)
        # global_ns=user_global_ns)
        # verbose=self.IP.rc.system_verbose)

        self.IP.raw_input = input_func
        sys.excepthook = excepthook
        self.iter_more = 0
        self.history_level = 0
        self.complete_sep = re.compile(r'[\s\{\}\[\]\(\)]')
        self.updateNamespace({'exit': lambda: None})
        self.updateNamespace({'quit': lambda: None})
        if int(IPYTHON_VERSION) < 5:  # HACK
            self.IP.readline_startup_hook(self.IP.pre_readline)
        # Workaround for updating namespace with sys.modules
        #
        self.__update_namespace()

        # help() is blocking, which hangs GTK+.
        import pydoc

        self.updateNamespace({'help': pydoc.doc})

        # Avoid using input splitter when not really needed.
        # Perhaps it could work even before 5.8.0
        # But it definitely does not work any more with >= 7.0.0
        self.no_input_splitter = IPYTHON_VERSION >= 5
        self.lines = []
        self.indent_spaces = ''

    def __update_namespace(self):
        """
        Update self.IP namespace for autocompletion with sys.modules
        """
        for key, value in sys.modules.items():
            if '.' not in key:
                self.IP.user_ns.update({key: value})

    def execute(self):
        """
        Executes the current line provided by the shell object.
        """
        self.history_level = 0

        # this is needed because some functions in IPython use 'print' to print
        # output (like 'who')
        #
        orig_stdout = sys.stdout
        sys.stdout = IPython.utils.io.stdout

        orig_stdin = sys.stdin
        sys.stdin = IPython.utils.io.stdin
        self.prompt = self.generatePrompt(self.iter_more)

        self.IP.hooks.pre_prompt_hook()
        if self.iter_more:
            try:
                self.prompt = self.generatePrompt(True)
            except:
                self.IP.showtraceback()
            if self.IP.autoindent:
                self.IP.rl_do_indent = True

        try:
            line = self.IP.raw_input(self.prompt)
        except KeyboardInterrupt:
            self.IP.write('\nKeyboardInterrupt\n')
            if self.no_input_splitter:
                self.lines = []
            else:
                self.IP.input_splitter.reset()
        except:
            self.IP.showtraceback()
        else:
            if self.no_input_splitter:
                self.lines.append(line)
                (status, self.indent_spaces) = self.IP.check_complete(
                    '\n'.join(self.lines)
                )
                self.iter_more = status == 'incomplete'
            else:
                self.IP.input_splitter.push(line)
                self.iter_more = self.IP.input_splitter.push_accepts_more()
            self.prompt = self.generatePrompt(self.iter_more)
            if not self.iter_more:
                if self.no_input_splitter:
                    source_raw = '\n'.join(self.lines)
                    self.lines = []
                else:
                    source_raw = self.IP.input_splitter.raw_reset()
                self.IP.run_cell(source_raw, store_history=True)
                self.IP.rl_do_indent = False
            else:
                # TODO: Auto-indent
                #
                self.IP.rl_do_indent = True

        sys.stdout = orig_stdout
        sys.stdin = orig_stdin

    def generatePrompt(self, is_continuation):
        """
        Generate prompt depending on is_continuation value

        @param is_continuation
        @type is_continuation: boolean

        @return: The prompt string representation
        @rtype: string

        """
        if IPYTHON_VERSION >= 5:  # HACK
            return '... ' if is_continuation else '>>> '
        if is_continuation:
            prompt = self.IP.prompt_manager.render('in2')
        else:
            prompt = self.IP.prompt_manager.render('in')

        return prompt

    def historyBack(self):
        """
        Provides one history command back.

        @return: The command string.
        @rtype: string
        """
        self.history_level -= 1
        if not self._getHistory():
            self.history_level += 1
        return self._getHistory()

    def historyForward(self):
        """
        Provides one history command forward.

        @return: The command string.
        @rtype: string
        """
        if self.history_level < 0:
            self.history_level += 1
        return self._getHistory()

    def _getHistory(self):
        """
        Get's the command string of the current history level.

        @return: Historic command string.
        @rtype: string
        """
        try:
            rv = self.IP.user_ns['In'][self.history_level].strip('\n')
        except IndexError:
            rv = ''
        return rv

    def updateNamespace(self, ns_dict):
        """
        Add the current dictionary to the shell namespace.

        @param ns_dict: A dictionary of symbol-values.
        @type ns_dict: dictionary
        """
        self.IP.user_ns.update(ns_dict)

    def complete(self, line):
        """
        Returns an auto completed line and/or possibilities for completion.

        @param line: Given line so far.
        @type line: string

        @return: Line completed as for as possible,
        and possible further completions.
        @rtype: tuple
        """
        split_line = self.complete_sep.split(line)
        if split_line[-1]:
            possibilities = self.IP.complete(split_line[-1])
        else:
            completed = line
            possibilities = ['', []]
        if possibilities:

            def _commonPrefix(str1, str2):
                """
                Reduction function. returns common prefix of two given strings.

                @param str1: First string.
                @type str1: string
                @param str2: Second string
                @type str2: string

                @return: Common prefix to both strings.
                @rtype: string
                """
                for i in range(len(str1)):
                    if not str2.startswith(str1[: i + 1]):
                        return str1[:i]
                return str1

            if possibilities[1]:
                common_prefix = (
                    reduce(_commonPrefix, possibilities[1]) or split_line[-1]
                )
                completed = line[: -len(split_line[-1])] + common_prefix
            else:
                completed = line
        else:
            completed = line
        return completed, possibilities[1]

    def shell(self, cmd, verbose=0, debug=0, header=''):
        """
        Replacement method to allow shell commands without them blocking.

        @param cmd: Shell command to execute.
        @type cmd: string
        @param verbose: Verbosity
        @type verbose: integer
        @param debug: Debug level
        @type debug: integer
        @param header: Header to be printed before output
        @type header: string
        """
        if verbose or debug:
            print((header + cmd))
        # flush stdout so we don't mangle python's buffering
        if not debug:
            popen = subprocess.Popen(
                cmd,
                bufsize=0,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                close_fds=True,
            )
            print((popen.stdout.read()))


class ConsoleView(Gtk.TextView):
    """
    Specialized text view for console-like workflow.

    @cvar ANSI_COLORS: Mapping of terminal colors to X11 names.
    @type ANSI_COLORS: dictionary

    @ivar text_buffer: Widget's text buffer.
    @type text_buffer: gtk.TextBuffer
    @ivar color_pat: Regex of terminal color pattern
    @type color_pat: _sre.SRE_Pattern
    @ivar mark: Scroll mark for automatic scrolling on input.
    @type mark: gtk.TextMark
    @ivar line_start: Start of command line mark.
    @type line_start: gtk.TextMark
    """

    ANSI_COLORS = {
        '0;30': 'Black',
        '0;31': 'Red',
        '0;32': 'Green',
        '0;33': 'Brown',
        '0;34': 'Blue',
        '0;35': 'Purple',
        '0;36': 'Cyan',
        '0;37': 'LightGray',
        '1;30': 'DarkGray',
        '1;31': 'DarkRed',
        '1;32': 'SeaGreen',
        '1;33': 'Yellow',
        '1;34': 'LightBlue',
        '1;35': 'MediumPurple',
        '1;36': 'LightCyan',
        '1;37': 'White',
    }

    def __init__(self):
        """
        Initialize console view.
        """
        Gtk.TextView.__init__(self)
        pango_ctx = self.get_pango_context()
        chosen = None
        for f in pango_ctx.list_families():
            name = f.get_name()
            # These are known to show e.g U+FFFC
            if name in ["Courier New", "Courier Mono"]:
                chosen = name
                break
            # if name in ["Liberation Sans"]:
            #    chosen = name
            #    # But prefer a monospace one if possible
        if chosen == None:
            chosen = "Mono"
        self.modify_font(Pango.FontDescription(chosen))
        self.set_cursor_visible(True)
        self.text_buffer = self.get_buffer()
        self.mark = self.text_buffer.create_mark(
            'scroll_mark', self.text_buffer.get_end_iter(), False
        )
        for code in self.ANSI_COLORS:
            self.text_buffer.create_tag(
                code, foreground=self.ANSI_COLORS[code], weight=700
            )
        self.text_buffer.create_tag('0')
        self.text_buffer.create_tag('notouch', editable=False)
        self.color_pat = re.compile(r'\x01?\x1b\[(.*?)m\x02?')
        self.line_start = self.text_buffer.create_mark(
            'line_start', self.text_buffer.get_end_iter(), True
        )
        self.connect('key-press-event', self.onKeyPress)

    def write(self, text, editable=False):
        GLib.idle_add(self._write, text, editable)

    def _write(self, text, editable=False):
        """
        Write given text to buffer.

        @param text: Text to append.
        @type text: string
        @param editable: If true, added text is editable.
        @type editable: boolean
        """
        segments = self.color_pat.split(text)
        segment = segments.pop(0)
        start_mark = self.text_buffer.create_mark(
            None, self.text_buffer.get_end_iter(), True
        )
        self.text_buffer.insert(self.text_buffer.get_end_iter(), segment)

        if segments:
            ansi_tags = self.color_pat.findall(text)
            for tag in ansi_tags:
                i = segments.index(tag)
                self.text_buffer.insert_with_tags_by_name(
                    self.text_buffer.get_end_iter(), segments[i + 1], tag
                )
                segments.pop(i)
        if not editable:
            self.text_buffer.apply_tag_by_name(
                'notouch',
                self.text_buffer.get_iter_at_mark(start_mark),
                self.text_buffer.get_end_iter(),
            )
        self.text_buffer.delete_mark(start_mark)
        self.scroll_mark_onscreen(self.mark)

    def showPrompt(self, prompt):
        GLib.idle_add(self._showPrompt, prompt)

    def _showPrompt(self, prompt):
        """
        Prints prompt at start of line.

        @param prompt: Prompt to print.
        @type prompt: string
        """
        self._write(prompt)
        self.text_buffer.move_mark(self.line_start, self.text_buffer.get_end_iter())

    def changeLine(self, text):
        GLib.idle_add(self._changeLine, text)

    def _changeLine(self, text):
        """
        Replace currently entered command line with given text.

        @param text: Text to use as replacement.
        @type text: string
        """
        text_iter = self.text_buffer.get_iter_at_mark(self.line_start)
        text_iter.forward_to_line_end()
        self.text_buffer.delete(
            self.text_buffer.get_iter_at_mark(self.line_start), text_iter
        )
        self._write(text, True)

    def getCurrentLine(self):
        """
        Get text in current command line.

        @return: Text of current command line.
        @rtype: string
        """
        rv = self.text_buffer.get_slice(
            self.text_buffer.get_iter_at_mark(self.line_start),
            self.text_buffer.get_end_iter(),
            False,
        )
        return rv

    def showReturned(self, text):
        GLib.idle_add(self._showReturned, text)

    def _showReturned(self, text):
        """
        Show returned text from last command and print new prompt.

        @param text: Text to show.
        @type text: string
        """
        text_iter = self.text_buffer.get_iter_at_mark(self.line_start)
        text_iter.forward_to_line_end()
        self.text_buffer.apply_tag_by_name(
            'notouch', self.text_buffer.get_iter_at_mark(self.line_start), text_iter
        )
        self._write('\n' + text)
        if text:
            self._write('\n')
        self._write('\n')  # Add extra line, like normal IPython
        self._showPrompt(self.prompt)
        self.text_buffer.move_mark(self.line_start, self.text_buffer.get_end_iter())
        self.text_buffer.place_cursor(self.text_buffer.get_end_iter())

        if self.IP.rl_do_indent:
            if self.no_input_splitter:
                indentation = self.indent_spaces
            else:
                indentation = self.IP.input_splitter.indent_spaces * ' '
            self.text_buffer.insert_at_cursor(indentation)

    def onKeyPress(self, _widget, event):
        """
        Key press callback used for correcting behavior for console-like
        interfaces. For example 'home' should go to prompt, not to beginning of
        line.

        @param widget: Widget that key press occurred in.
        @type widget: gtk.Widget
        @param event: Event object
        @type event: gtk.gdk.Event

        @return: Return True if event should not trickle.
        @rtype: boolean
        """
        insert_mark = self.text_buffer.get_insert()
        insert_iter = self.text_buffer.get_iter_at_mark(insert_mark)
        selection_mark = self.text_buffer.get_selection_bound()
        selection_iter = self.text_buffer.get_iter_at_mark(selection_mark)
        start_iter = self.text_buffer.get_iter_at_mark(self.line_start)
        if event.keyval == Gdk.KEY_Home:
            if (
                event.state & Gdk.ModifierType.CONTROL_MASK
                or event.state & Gdk.ModifierType.MOD1_MASK
            ):
                pass
            elif event.state & Gdk.ModifierType.SHIFT_MASK:
                self.text_buffer.move_mark(insert_mark, start_iter)
                return True
            else:
                self.text_buffer.place_cursor(start_iter)
                return True
        elif event.keyval == Gdk.KEY_Left:
            insert_iter.backward_cursor_position()
            if not insert_iter.editable(True):
                return True
        elif not event.string:
            pass
        elif (
            start_iter.compare(insert_iter) <= 0
            and start_iter.compare(selection_iter) <= 0
        ):
            pass
        elif (
            start_iter.compare(insert_iter) > 0
            and start_iter.compare(selection_iter) > 0
        ):
            self.text_buffer.place_cursor(start_iter)
        elif insert_iter.compare(selection_iter) < 0:
            self.text_buffer.move_mark(insert_mark, start_iter)
        elif insert_iter.compare(selection_iter) > 0:
            self.text_buffer.move_mark(selection_mark, start_iter)

        return self.onKeyPressExtend(event)

    def onKeyPressExtend(self, event):
        """
        For some reason we can't extend onKeyPress directly (bug #500900).
        """
        pass


class IPythonView(ConsoleView, IterableIPShell):
    """
    Sub-class of both modified IPython shell and L{ConsoleView} this makes
    a GTK+ IPython console.
    """

    def __init__(self):
        """
        Initialize. Redirect I/O to console.
        """
        ConsoleView.__init__(self)
        self.cout = StringIO()
        IterableIPShell.__init__(
            self, cout=self.cout, cerr=self.cout, input_func=self.raw_input
        )
        #        self.connect('key_press_event', self.keyPress)
        self.interrupt = False
        self.execute()
        self.prompt = self.generatePrompt(False)
        self.cout.truncate(0)
        self.showPrompt(self.prompt)

    def raw_input(self, prompt=''):
        """
        Custom raw_input() replacement. Get's current line from console buffer.

        @param prompt: Prompt to print. Here for compatibility as replacement.
        @type prompt: string

        @return: The current command line text.
        @rtype: string
        """
        if self.interrupt:
            self.interrupt = False
            raise KeyboardInterrupt
        return self.getCurrentLine()

    def onKeyPressExtend(self, event):
        """
        Key press callback with plenty of shell goodness, like history,
        autocompletions, etc.

        @param widget: Widget that key press occurred in.
        @type widget: gtk.Widget
        @param event: Event object.
        @type event: gtk.gdk.Event

        @return: True if event should not trickle.
        @rtype: boolean
        """
        if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == 99:
            self.interrupt = True
            self._processLine()
            return True
        elif event.keyval == Gdk.KEY_Return:
            self._processLine()
            return True
        elif event.keyval == Gdk.KEY_Up:
            self.changeLine(self.historyBack())
            return True
        elif event.keyval == Gdk.KEY_Down:
            self.changeLine(self.historyForward())
            return True
        elif event.keyval == Gdk.KEY_Tab:
            if not self.getCurrentLine().strip():
                return False
            completed, possibilities = self.complete(self.getCurrentLine())
            if len(possibilities) > 1:
                line_slice = self.getCurrentLine()
                self.write('\n')
                for symbol in possibilities:
                    self.write(symbol + '\n')
                self.showPrompt(self.prompt)
            self.changeLine(completed or line_slice)
            return True

    def _processLine(self):
        """
        Process current command line.
        """
        self.history_pos = 0
        self.execute()
        returnvalue = self.cout.getvalue()
        if returnvalue:
            returnvalue = returnvalue.strip('\n')
        self.showReturned(returnvalue)
        self.cout.truncate(0)
        self.cout.seek(0)


if __name__ == "__main__":
    window = Gtk.Window()
    window.set_default_size(640, 320)
    window.connect('delete-event', lambda x, y: Gtk.main_quit())
    window.add(IPythonView())
    window.show_all()
    Gtk.main()



# Copyright (C) 2009-2010
#     Adam Olsen <arolsen@gmail.com>
#     Abhishek Mukherjee <abhishek.mukher.g@gmail.com>
#     Steve Dodier <sidnioulzg@gmail.com>
# Copyright (C) 2017 Christian Stadelmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import html
import logging

import gi
from gi.repository import Gtk, GLib

from xl.player.adapters import PlaybackAdapter
from xl import covers, common
from xl import event as xl_event
from xl import player as xl_player
from xl import settings as xl_settings
from xl.nls import gettext as _
from xlgui import icons
from xlgui.guiutil import pixbuf_from_data

from . import notifyprefs


# For documentation on libnotify see also the "Desktop Notifications Specification":
# https://developer.gnome.org/notification-spec/
gi.require_version('Notify', '0.7')
from gi.repository import Notify


LOGGER = logging.getLogger(__name__)
DEFAULT_ICON_SIZE = (48, 48)

BODY_ARTIST_ALBUM = _('from {album} by {artist}')
BODY_ARTIST = _('by {artist}')
BODY_ALBUM = _('by {album}')


class NotifierSettings:
    @staticmethod
    def __inner_preference(klass):
        """Function will make a property for a given subclass of Preference"""

        def getter(self):
            return xl_settings.get_option(klass.name, klass.default or None)

        def setter(self, val):
            xl_settings.set_option(klass.name, val)

        # Migration:
        if hasattr(klass, 'pre_migration_name'):
            if xl_settings.get_option(klass.name, None) is None:
                old_value = xl_settings.get_option(
                    klass.pre_migration_name, klass.default or None
                )
                if old_value is not None:
                    xl_settings.set_option(klass.name, old_value)
                    xl_settings.MANAGER.remove_option(klass.name)

        return property(getter, setter)

    notify_pause = __inner_preference(notifyprefs.NotifyPause)
    resize_covers = __inner_preference(notifyprefs.ResizeCovers)
    show_covers = __inner_preference(notifyprefs.ShowCovers)
    show_when_focused = __inner_preference(notifyprefs.ShowWhenFocused)
    tray_hover = __inner_preference(notifyprefs.TrayHover)
    use_media_icons = __inner_preference(notifyprefs.UseMediaIcons)


class Notifier(PlaybackAdapter):

    settings = NotifierSettings()

    def __init__(self, exaile, caps):
        self.__exaile = exaile
        self.__old_icon = None
        self.__tray_connection = None

        notification = Notify.Notification.new("", None, None)

        if 'sound' in caps:
            notification.set_hint('suppress-sound', GLib.Variant.new_boolean(True))
        self.settings.can_show_markup = 'body-markup' in caps

        notification.set_urgency(Notify.Urgency.LOW)
        notification.set_timeout(Notify.EXPIRES_DEFAULT)

        # default action is invoked on clicking the notification
        notification.add_action(
            "default", "this should never be displayed", self.__on_default_action
        )

        self.notification = notification
        PlaybackAdapter.__init__(self, xl_player.PLAYER)

        xl_event.add_callback(self.on_option_set, 'plugin_notify_option_set')
        xl_event.add_callback(self.on_option_set, 'gui_option_set')
        # initial setup through options:
        self.on_option_set(None, xl_settings, notifyprefs.TrayHover.name)
        self.on_option_set(None, xl_settings, notifyprefs.ShowCovers.name)

    def __on_default_action(self, _notification, _action):
        self.__exaile.gui.main.window.present()

    def on_playback_player_end(self, _event, player, track):
        if self.settings.notify_pause:
            self.update_track_notify(player, track, 'media-playback-stop')

    def on_playback_track_start(self, _event, player, track):
        self.update_track_notify(player, track)

    def on_playback_toggle_pause(self, _event, player, track):
        if self.settings.notify_pause:
            if player.is_paused():
                self.update_track_notify(player, track, 'media-playback-pause')
            else:
                self.update_track_notify(player, track, 'media-playback-start')

    def on_playback_error(self, _event, _player, message):
        self.__maybe_show_notification("Playback error", message, 'dialog-error')

    def __on_query_tooltip(self, *_args):
        if self.settings.tray_hover:
            self.__maybe_show_notification(force_show=True)

    @common.idle_add()
    def __try_connect_tray(self):
        tray_icon = self.__exaile.gui.tray_icon
        if tray_icon:
            self.__tray_connection = tray_icon.connect(
                'query-tooltip', self.__on_query_tooltip
            )
        else:
            LOGGER.warning("Tried to connect to non-existing tray icon")

    def __set_tray_hover_state(self, state):
        # Interacting with the tray might break if our option handler is being
        # invoked when exaile.gui.tray_icon == None. This might happen because
        # it is not defined which 'option_set' handler is being invoked first.
        tray_icon = self.__exaile.gui.tray_icon
        if state and not self.__tray_connection:
            if tray_icon:
                self.__tray_connection = tray_icon.connect(
                    'query-tooltip', self.__on_query_tooltip
                )
            else:
                self.__try_connect_tray()
        elif not state and self.__tray_connection:
            if tray_icon:  # xlgui.main might already have destroyed the tray icon
                self.__exaile.gui.tray_icon.disconnect(self.__tray_connection)
            self.__tray_connection = None

    def on_option_set(self, _event, settings, option):
        if option == notifyprefs.TrayHover.name or option == 'gui/use_tray':
            has_tray = settings.get_option('gui/use_tray')
            shall_show_tray_hover = self.settings.tray_hover and has_tray
            self.__set_tray_hover_state(shall_show_tray_hover)
        elif option == notifyprefs.ShowCovers.name:
            # delete current cover if user doesn't want to see more covers
            if not self.settings.show_covers:
                if self.settings.resize_covers:
                    size = DEFAULT_ICON_SIZE[1]
                else:
                    size = Gtk.IconSize.DIALOG
                new_icon = icons.MANAGER.pixbuf_from_icon_name('exaile', size)
                self.notification.set_image_from_pixbuf(new_icon)

    def __maybe_show_notification(
        self, summary=None, body='', icon_name=None, force_show=False
    ):
        # If summary is none, don't update the Notification
        if summary is not None:
            try:
                self.notification.update(summary, body, icon_name)
            except GLib.Error:
                LOGGER.exception("Could not set new notification status.")
                return
        # decide whether to show the notification or not
        if (
            self.settings.show_when_focused
            or not self.__exaile.gui.main.window.is_active()
            or force_show
        ):
            try:
                self.notification.show()
            except GLib.Error:
                LOGGER.exception("Could not set new notification status.")
                self.__exaile.plugins.disable_plugin(__name__)

    def __get_body_str(self, track):
        artist_str = html.escape(
            track.get_tag_display('artist', artist_compilations=False)
        )
        album_str = html.escape(track.get_tag_display('album'))

        if self.settings.can_show_markup:
            if artist_str:
                artist_str = '<i>%s</i>' % artist_str
            if album_str:
                album_str = '<i>%s</i>' % album_str

        if artist_str and album_str:
            body = BODY_ARTIST_ALBUM.format(artist=artist_str, album=album_str)
        elif artist_str:
            body = BODY_ARTIST.format(artist=artist_str)
        elif album_str:
            body = BODY_ALBUM.format(album=album_str)
        else:
            body = ""
        return body

    def __get_icon(self, track, media_icon):
        # TODO: icons are too small, even with settings.resize_covers=False
        icon_name = None
        if media_icon and self.settings.use_media_icons:
            icon_name = media_icon
        elif self.settings.show_covers:
            cover_data = covers.MANAGER.get_cover(
                track, set_only=True, use_default=True
            )
            size = DEFAULT_ICON_SIZE if self.settings.resize_covers else None
            new_icon = pixbuf_from_data(cover_data, size)
            self.notification.set_image_from_pixbuf(new_icon)
        return icon_name

    def update_track_notify(self, _player, track, media_icon=None):
        # TODO: notification.add_action(): previous, play/pause, next ?
        title = html.escape(track.get_tag_display('title'))

        summary = title
        body = self.__get_body_str(track)
        icon_name = self.__get_icon(track, media_icon)

        # If icon_name is None, the previous icon will not be replaced
        self.__maybe_show_notification(summary, body, icon_name)

    def destroy(self):
        PlaybackAdapter.destroy(self)
        self.__set_tray_hover_state(False)
        notification = self.notification
        # must be called on separate thread, since it is a synchronous call and might block
        self.__close_notification(notification)
        self.notification.clear_actions()
        self.notification = None
        self.__exaile = None

    @staticmethod
    @common.threaded
    def __close_notification(notification):
        try:
            notification.close()
        except GLib.Error:
            LOGGER.exception("Failed to close notification")


class NotifyPlugin:
    def __init__(self):
        self.__notifier = None
        self.__exaile = None

    def enable(self, exaile):
        self.__exaile = exaile
        self.__init_notify()

    @common.threaded
    def __init_notify(self):
        can_continue = True
        caps = None
        if not Notify.is_initted():
            can_continue = Notify.init('Exaile')
        if not can_continue:
            LOGGER.error("Notify.init() returned false.")

        if can_continue:
            # This is the first synchronous call to the Notify server.
            # This call might fail if no server is present or it is broken.
            # Test it on window manager sessions (e.g. Weston) without
            # libnotify support, not on a Desktop Environment (such as
            # GNOME, KDE) to reproduce.
            available, name, vendor, version, spec_version = Notify.get_server_info()
            if available:
                LOGGER.info(
                    "Connected with notify server %s (version %s) by %s",
                    name,
                    version,
                    vendor,
                )
                LOGGER.info("Supported spec version: %s", spec_version)
                # This is another synchronous, blocking call:
                caps = Notify.get_server_caps()
                # Example from Fedora 26 Linux with GNOME on Wayland:
                # ['actions', 'body', 'body-markup', 'icon-static', 'persistence', 'sound']
                LOGGER.debug("Notify server caps: %s", caps)
                if not caps or not isinstance(caps, list):
                    can_continue = False
            else:
                LOGGER.error(
                    "Failed to retrieve capabilities from notify server. "
                    "This may happen if the desktop environment does not support "
                    "the org.freedesktop.Notifications DBus interface."
                )
                can_continue = False
        self.__handle_init(can_continue, caps)

    # Must be run on main thread because we need to make sure that the plugin
    # is not being disabled while this function runs. Otherwise, race
    # conditions might trigger obscure bugs.
    @common.idle_add()
    def __handle_init(self, can_continue, caps):
        exaile = self.__exaile
        if exaile is None:  # Plugin has been disabled in the mean time
            return

        if can_continue:  # check again, might have changed
            if exaile.loading:
                xl_event.add_ui_callback(self.__init_notifier, 'gui_loaded', None, caps)
            else:
                self.__init_notifier(caps)
        else:
            LOGGER.warning("Disabling NotifyPlugin.")
            exaile.plugins.disable_plugin(__name__)
            if not exaile.loading:
                # TODO: send error to GUI
                pass

    def __init_notifier(self, caps):
        self.__notifier = Notifier(self.__exaile, caps)
        return GLib.SOURCE_REMOVE

    def disable(self, exaile):
        self.teardown(exaile)

        self.__exaile = None

    def teardown(self, exaile):
        if self.__notifier:
            self.__notifier.destroy()
            self.__notifier = None

    def get_preferences_pane(self):
        return notifyprefs


plugin_class = NotifyPlugin
